# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from decimal import Decimal, ROUND_CEILING

import filters as f
from filters.test import BaseFilterTestCase


class DecimalTestCase(BaseFilterTestCase):
    filter_type = f.Decimal

    def test_pass_none(self):
        """
        `None` always passes this Filter.

        Use `Required | Decimal` if you want to reject `None`.
        """
        self.assertFilterPasses(None)

    def test_pass_valid_decimal(self):
        """
        The incoming value can be interpreted as a Decimal.
        """
        value = '3.1415926'

        self.assertFilterPasses(value, Decimal(value))

    def test_pass_max_precision(self):
        """
        You can limit the max precision of the resulting Decimal
        object.

        Note that a too-precise value is not considered invalid;
        instead, it gets rounded to the expected precision.
        """
        self.assertFilterPasses(
            self._filter('3.1415926', max_precision=3),
            Decimal('3.142'),
        )

    def test_max_precision_quantized(self):
        """
        ``max_precision`` can also be specified as a Decimal object.
        """
        self.assertFilterPasses(
            self._filter('3.1415926', max_precision=Decimal('0.001')),
            Decimal('3.142'),
        )

    def test_pass_zero(self):
        """
        0 is also considered a valid Decimal value.
        """
        value = '0'
        self.assertFilterPasses(value, Decimal(value))

    def test_pass_scientific_notation(self):
        """
        Scientific notation is considered valid, as in certain cases it
        may be the only way to represent a really large or small
        value.
        """
        value = '2.8E-12'
        self.assertFilterPasses(value, Decimal(value))

    def test_pass_boolean(self):
        """
        Booleans are technically ints, so they can be converted to
        Decimals.
        """
        value = True
        self.assertFilterPasses(value, Decimal(value))

    def test_fail_invalid_value(self):
        """
        The incoming value cannot be converted to a Decimal.
        """
        self.assertFilterErrors(
            'this is not a decimal',
            [f.Decimal.CODE_INVALID],
        )

    def test_fail_non_finite_value(self):
        """
        Non-finite values like 'NaN' and 'Inf' are considered invalid,
        even though they are technically parseable.
        """
        self.assertFilterErrors( 'NaN', [f.Decimal.CODE_NON_FINITE])
        self.assertFilterErrors('+Inf', [f.Decimal.CODE_NON_FINITE])
        self.assertFilterErrors('-Inf', [f.Decimal.CODE_NON_FINITE])
        # There are a few other possible non-finite values out there,
        # but you get the idea.

    def test_pass_tuple(self):
        """
        You may pass a 3-tuple for more control over how the resulting
        Decimal object is created.
        """
        value = (0, (4, 2), -1)
        self.assertFilterPasses(value, Decimal(value))

    def test_fail_tuple_invalid(self):
        """
        If you're going to use a tuple, you've got to make sure you get
        it right!
        """
        self.assertFilterErrors(('1', '2', '3'), [
          f.Decimal.CODE_INVALID])

    def test_fail_tuple_disallowed(self):
        """
        The filter is configured to disallow tuple values.
        """
        self.assertFilterErrors(
            self._filter((0, (4, 2), -1), allow_tuples=False),
            [f.Type.CODE_WRONG_TYPE],
        )

    def test_fail_bytes(self):
        """
        To ensure that the filter behaves the same in Python 2 and
        Python 3, bytes are not allowed.
        """
        self.assertFilterErrors(b'-12', [f.Type.CODE_WRONG_TYPE])

    def test_fail_unsupported_type(self):
        """
        The incoming value has an unsupported type.
        """
        self.assertFilterErrors(
            {0, (4, 2), -1},
            [f.Type.CODE_WRONG_TYPE],
        )


class IntTestCase(BaseFilterTestCase):
    filter_type = f.Int

    def test_pass_none(self):
        """
        `None` always passes this Filter.

        Use `Required | Int` if you want to reject `None`.
        """
        self.assertFilterPasses(None)

    def test_pass_valid_int(self):
        """
        The incoming value can be interpreted as an int.
        """
        self.assertFilterPasses('42', 42)

    def test_pass_zero(self):
        """
        The incoming value is zero.
        """
        self.assertFilterPasses('0', 0)

    def test_pass_negative(self):
        """
        The incoming value is a negative int.
        """
        self.assertFilterPasses('-314', -314)

    def test_pass_boolean(self):
        """
        Booleans are technically ints.
        """
        self.assertFilterPasses(True, 1)

    def test_fail_invalid_value(self):
        """
        The incoming value cannot be interpreted as a number, let alone
        an integer.
        """
        self.assertFilterErrors(
            'this is not an int',
            [f.Decimal.CODE_INVALID],
        )

    def test_fail_bytes(self):
        """
        To ensure that the filter behaves the same in Python 2 and
        Python 3, bytes are not allowed.
        """
        self.assertFilterErrors(b'-12', [f.Type.CODE_WRONG_TYPE])

    def test_fail_float_value(self):
        """
        The incoming value contains significant digits after the
        decimal point.
        """
        self.assertFilterErrors(
            '42.01',
            [f.Int.CODE_DECIMAL],
        )

    def test_pass_int_point_zero(self):
        """
        The incoming value contains only insignificant digits after the
        decimal point.
        """
        self.assertFilterPasses('42.0000000000000', 42)

    def test_pass_scientific_notation(self):
        """
        The incoming value is expressed in scientific notation.
        """
        self.assertFilterPasses('2.6E4', 26000)

    def test_fail_non_finite_value(self):
        """
        The incoming value is a non-finite value.
        """
        self.assertFilterErrors( 'NaN', [f.Decimal.CODE_NON_FINITE])
        self.assertFilterErrors('+Inf', [f.Decimal.CODE_NON_FINITE])
        self.assertFilterErrors('-Inf', [f.Decimal.CODE_NON_FINITE])
        # There are a few other possible non-finite values out there,
        # but you get the idea.

    def test_pass_int(self):
        """
        The incoming value is already an int object.
        """
        self.assertFilterPasses(777)

class MaxTestCase(BaseFilterTestCase):
    filter_type = f.Max

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use ``Required | Max`` if you want to reject ``None``.
        """
        self.assertFilterPasses(
            self._filter(None, max_value=5),
        )

    def test_pass_lesser_value(self):
        """
        The incoming value is smaller than the max value.
        """
        self.assertFilterPasses(
            self._filter(3, max_value=5),
        )

    def test_pass_equal_value(self):
        """
        The incoming value is equal to the max value.
        """
        self.assertFilterPasses(
            self._filter(5, max_value=5),
        )

    def test_fail_equal_value_exclusive_comparison(self):
        """
        The incoming value is equal to the max value, but the Filter is
        configured to use an exclusive comparison.

        This is useful for infinite-precision floats and other cases
        where it is impossible to specify the max value exactly.
        """
        self.assertFilterErrors(
            # The Filter is configured to allow any float value that is
            # less than - but not equal to - 5.0.
            self._filter(5.0, max_value=5.0, exclusive=True),
            [f.Max.CODE_TOO_BIG],
        )

    def test_fail_greater_value(self):
        """
        The incoming value is greater than the max value.
        """
        self.assertFilterErrors(
            self._filter(8, max_value=5),
            [f.Max.CODE_TOO_BIG],
        )

    def test_string_comparison_oddness(self):
        """
        If the filter is being used on strings, the comparison is case
        sensitive.

        Note:  due to the way ASCII works, this may yield unexpected
        results (lowercase > uppercase).  Also, watch out for
        Unicode oddness!

        Basically what I'm trying to say is, don't use this filter on
        string values.
        """
        # ord('F') => 70
        # ord('f') => 102
        self.assertFilterErrors(
            self._filter('foo', max_value='Foo'),
            [f.Max.CODE_TOO_BIG],
        )


class MinTestCase(BaseFilterTestCase):
    filter_type = f.Min

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use ``Required | Min`` if you want to reject ``None``.
        """
        self.assertFilterPasses(
            self._filter(None, min_value=5),
        )

    def test_pass_greater_value(self):
        """
        The incoming value is greater than the min value.
        """
        self.assertFilterPasses(
            self._filter(8, min_value=5),
        )

    def test_pass_equal_value(self):
        """
        The incoming value is equal to the min value.
        """
        self.assertFilterPasses(
            self._filter(5, min_value=5),
        )

    def test_fail_equal_value_exclusive_comparison(self):
        """
        The incoming value is equal to the min value, but the Filter is
        configured to use exclusive comparison.

        This is useful for infinite-precision floats and other cases
        where it is impossible to specify the min value exactly.
        """
        self.assertFilterErrors(
            # The Filter is configured to allow any float value that is
            #   greater than - but not equal to - 5.0.
            self._filter(5.0, min_value=5.0, exclusive=True),
            [f.Min.CODE_TOO_SMALL],
        )

    def test_fail_lesser_value(self):
        """
        The incoming value is less than the min value.
        """
        self.assertFilterErrors(
            self._filter(4, min_value=5),
            [f.Min.CODE_TOO_SMALL],
        )

    def test_string_comparison_oddness(self):
        """
        If the filter is being used on strings, the comparison is case
        sensitive.

        Note:  due to the way ASCII works, this may yield unexpected
        results (lowercase > uppercase).  Also, watch out for
        Unicode oddness!

        Basically what I'm trying to say is, don't use this filter on
        string values.
        """
        # ord('f') => 102
        # ord('F') => 70
        self.assertFilterErrors(
            self._filter('Foo', min_value='foo'),
            [f.Min.CODE_TOO_SMALL],
        )


class RoundTestCase(BaseFilterTestCase):
    filter_type = f.Round

    def test_pass_none(self):
        """
        `None` always passes this filter.

        Use `Required | Round` to reject incoming `None`.
        """
        self.assertFilterPasses(None)

    def test_pass_round_integer_to_nearest_integer(self):
        """
        Rounds an integer to the nearest integer value.
        """
        self.assertFilterPasses(
            # You should always specify `to_nearest` as a string, to
            # avoid floating point issues.
            self._filter(42, to_nearest='5'),

            # The result is always a Decimal object.
            Decimal('40.0'),
        )

    def test_pass_round_integer_to_nearest_float(self):
        """
        Rounds an integer to the nearest float value.
        """
        self.assertFilterPasses(
            self._filter(42, to_nearest='5.5'),

            Decimal('44.0'),
        )

    def test_pass_round_float_to_nearest_integer(self):
        """
        Rounds a float to the nearest integer value.
        """
        self.assertFilterPasses(
            self._filter(3.5, to_nearest='1'),
            Decimal('4.0'),
        )

    def test_pass_round_float_to_nearest_float(self):
        """
        Rounds a float to the nearest float value.
        """
        # Just to be tricky, use a float value that would normally
        # result in some nasty floating point artifacts.
        # http://stackoverflow.com/a/4340355
        self.assertFilterPasses(
            self._filter(1.368161685161, to_nearest='0.05'),
            Decimal('1.35'),
        )

    def test_pass_round_string_float(self):
        """
        Rounds a float represented as a string to avoid floating point
        issues.
        """
        # http://stackoverflow.com/q/22599883
        self.assertFilterPasses(
            self._filter('2.775', to_nearest='0.1'),
            Decimal('2.8'),
        )

    def test_pass_round_to_big_value(self):
        """
        Rounds something to a value greater than 1.
        """
        self.assertFilterPasses(
            self._filter('386.428', to_nearest='20'),
            Decimal('380'),
        )

    def test_pass_round_negative_value(self):
        """
        Rounds a negative value.
        """
        self.assertFilterPasses(
            self._filter('-2.775', to_nearest='0.1'),
            Decimal('-2.8'),
        )

    def test_pass_modify_rounding(self):
        """
        By default, the filter will round up any value that is halfway
        to the nearest `to_nearest` value, but this behavior can be
        customized.
        """
        self.assertFilterPasses(
            self._filter('0.00000000001', rounding=ROUND_CEILING),
            Decimal('1'),
        )

    def test_pass_custom_result_type(self):
        """
        You can customize the return type of the filter.
        """
        self.assertFilterPasses(
            self._filter('2.775', result_type=int),
            3,
        )

    def test_fail_wrong_type(self):
        """
        The incoming value is not numeric.
        """
        self.assertFilterErrors('three', [f.Decimal.CODE_INVALID])



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from collections import OrderedDict

import filters as f
from filters.test import BaseFilterTestCase


class FilterChainTestCase(BaseFilterTestCase):
    def test_implicit_chain(self):
        """
        Chaining two filters together creates a FilterChain.
        """
        self.filter_type = lambda: f.Int | f.Max(3)

        self.assertFilterPasses('1', 1)
        self.assertFilterErrors('4', [f.Max.CODE_TOO_BIG])

    def test_implicit_chain_null(self):
        """
        Chaining a filter with ``None`` also yields a FilterChain, but
        unsurprisingly, the chain only contains the one filter.
        """
        filter_chain = f.Int() | None
        self.assertIsInstance(filter_chain, f.FilterChain)

        with self.assertRaises(f.FilterError):
            filter_chain.apply('not an int')

    # noinspection SpellCheckingInspection
    def test_chainception(self):
        """
        You can also chain FilterChains together.
        """
        fc1 = f.NotEmpty | f.Choice(choices=('Lucky', 'Dusty', 'Ned'))
        fc2 = f.NotEmpty | f.MinLength(4)

        self.filter_type = lambda: fc1 | fc2

        self.assertFilterPasses('Lucky')
        self.assertFilterErrors('El Guapo', [f.Choice.CODE_INVALID])
        self.assertFilterErrors('Ned', [f.MinLength.CODE_TOO_SHORT])

    def test_stop_after_invalid_value(self):
        """
        A FilterChain stops processing the incoming value after any
        filter fails.
        """
        # This FilterChain will pretty much reject anything that you
        # throw at it.
        self.filter_type =\
            lambda: f.MaxLength(3) | f.MinLength(8) | f.Required

        # Note that the value 'foobar' fails both the MaxLength and the
        # MinLength filters, but the FilterChain stops processing
        # after MaxLength fails.
        self.assertFilterErrors('foobar', [f.MaxLength.CODE_TOO_LONG])


class FilterRepeaterTestCase(BaseFilterTestCase):
    def test_pass_none(self):
        """
        For consistency with all the other Filter classes, `None` is
        considered a valid value to pass to a FilterRepeater, even
        though it is not iterable.
        """
        self.filter_type = lambda: f.FilterRepeater(f.Int)

        self.assertFilterPasses(None)

    def test_pass_iterable(self):
        """
        A FilterRepeater is applied to a list of valid values.
        """
        self.filter_type = lambda: f.FilterRepeater(f.NotEmpty | f.Int)

        self.assertFilterPasses(
            ['1', 2, 0, None, '-12'],
            [1, 2, 0, None, -12],
        )

    def test_fail_iterable(self):
        """
        A FilterRepeater is applied to a list that contains invalid
        values.
        """
        self.filter_type = lambda: f.FilterRepeater(f.NotEmpty | f.Int)

        self.assertFilterErrors(
            # First element is valid (control group).
            # The rest fail miserably.
            [4, 'NaN', 3.14, 'FOO', ''],

            {
                '1':    [f.Decimal.CODE_NON_FINITE],
                '2':    [f.Int.CODE_DECIMAL],
                '3':    [f.Decimal.CODE_INVALID],
                '4':    [f.NotEmpty.CODE_EMPTY],
            },

            expected_value = [4, None, None, None, None],
        )

    def test_pass_mapping(self):
        """
        A FilterRepeater is applied to a dict of valid values.
        """
        self.filter_type = lambda: f.FilterRepeater(f.NotEmpty | f.Int)

        self.assertFilterPasses(
            {
                'foo':      '1',
                'bar':      2,
                'baz':      None,
                'luhrmann': '-12',
            },

            # The FilterRepeater applies the filter chain to the dict's
            # values.  Note that it completely ignores the keys.
            {
                'foo':      1,
                'bar':      2,
                'baz':      None,
                'luhrmann': -12,
            },
        )

    def test_fail_mapping(self):
        """
        A FilterRepeater is applied to a dict that contains invalid
        values.
        """
        self.filter_type = lambda: f.FilterRepeater(f.NotEmpty | f.Int)

        self.assertFilterErrors(
            {
                # First element is valid (control group).
                # The rest fail miserably.
                'foo':      4,
                'bar':      'NaN',
                'baz':      3.14,
                'luhrmann': 'FOO',
            },

            {
                'bar':      [f.Decimal.CODE_NON_FINITE],
                'baz':      [f.Int.CODE_DECIMAL],
                'luhrmann': [f.Decimal.CODE_INVALID],
            },

            # Just as with collections, the invalid values in the
            # filtered value are replaced with `None`.
            expected_value = {
                'foo':      4,
                'bar':      None,
                'baz':      None,
                'luhrmann': None,
            },
        )

    def test_restrict_keys(self):
        """
        A FilterRepeated is configured to restrict allowed keys in a
        mapping.
        """
        self.filter_type = lambda: f.FilterRepeater(
            filter_chain    =f.Int,
            restrict_keys   = {'ducks', 'sea otters'},
        )

        # As long as you stick to the expected keys, everything's
        # hunky-dory.
        self.assertFilterPasses(
            {'ducks': '3', 'sea otters': '4'},
            {'ducks': 3, 'sea otters': 4},
        )

        # However, should you deviate from the marked path, we cannot
        # be held responsible for the consequences.
        self.assertFilterErrors(
            # Charlie shot first!
            {'ducks': '3', 'hawks': '4'},

            {
                'hawks':    [f.FilterRepeater.CODE_EXTRA_KEY],
            },

            # Invalid keys are not included in the filtered value.
            # This is very similar to how FilterMapper works.
            expected_value = {
                'ducks':    3,
            }
        )

    def test_restrict_indexes(self):
        """
        A FilterRepeater CAN be configured to restrict keys for
        incoming Iterables, although it is probably the wrong tool
        for the job (MaxLength is probably a better fit).
        """
        #
        # Note that if `restrict_keys` contains non-integers and/or
        # starts with a value other than 0, the FilterRepeater will
        # reject EVERY Iterable it comes across!
        #
        # Really, you should just stick a MaxLength(2) in front of the
        # FilterRepeater and call it a day.  It's less likely to
        # introduce a logic bug and way easier for other devs to
        # interpret.
        #
        # noinspection PyTypeChecker
        self.filter_type = lambda: f.FilterRepeater(
            filter_chain    =f.Int,
            restrict_keys   = {0, 1, 3, 4},
        )

        self.assertFilterPasses(['4', '3'], [4, 3])

        self.assertFilterErrors(
            ['50', '40', '30', '20', '10'],

            {
                # Index 2 was unexpected (the Filter is configured
                # only to allow indexes 0, 1, 3 and 4).
                '2': [f.FilterRepeater.CODE_EXTRA_KEY],
            },

            # To make things even more confusing, the invalid "keys"
            # (indexes) ARE included in the filtered value.  This is
            # because, unlike in mappings, it is not possible to
            # identify "missing" indexes.
            expected_value = [50, 40, None, 20, 10]
        )

        # The moral of the story is, don't use `restrict_keys` when
        # configuring a FilterRepeater that will operate on
        # collections.

    def test_fail_non_iterable_value(self):
        """
        A FilterRepeater will reject any non-iterable value it comes
        across (except for `None`).
        """
        self.filter_type = lambda: f.FilterRepeater(f.Int)

        self.assertFilterErrors(42, [f.Type.CODE_WRONG_TYPE])

    def test_repeater_chained_with_repeater(self):
        """
        Chaining two FilterRepeaters together has basically the same
        effect as combining their Filters, except for one very
        important difference:  The two sets of Filters are applied
        in sequence.

        That is, the second set of Filters only get applied if ALL
        of the Filters in the first set pass!

        Generally, combining two FilterRepeaters into a single instance
        is much easier to read/maintain than chaining them, but
        should you ever come across a situation where you need to
        apply two FilterRepeaters in sequence, you can do so.
        """
        self.filter_type =\
            lambda: f.FilterRepeater(f.NotEmpty) | f.FilterRepeater(
              f.Int)

        # The values in this list pass through both FilterRepeaters
        # successfully.
        self.assertFilterPasses(
            ['1', 2, 0, None, '-12'],
            [1, 2, 0, None, -12],
        )

        # The values in this list fail one or more Filters in each
        # FilterRepeater.
        self.assertFilterErrors(
            ['', 'NaN', 0, None, 'FOO'],

            {
                # Fails the NotEmpty filter in the first FilterRepeater.
                '0': [f.NotEmpty.CODE_EMPTY],

                # IMPORTANT:  Because the first FilterRepeater had one
                # or more errors, the outer FilterChain stopped.
                # # Fails the Int filter in the second FilterRepeater.
                # '1': [f.Decimal.CODE_NON_FINITE],
                # '4': [f.Int.CODE_INVALID],
            },

            # The result is the same as if we only ran the value
            # through the first FilterRepeater.
            expected_value = [
                None, 'NaN', 0, None, 'FOO'
            ]
        )

        # The values in this list pass the first FilterRepeater but
        # fail the second one.
        self.assertFilterErrors(
            ['1', 'NaN', 0, None, 'FOO'],

            {
                '1': [f.Decimal.CODE_NON_FINITE],
                '4': [f.Decimal.CODE_INVALID],
            },

            expected_value = [1, None, 0, None, None],
        )

    def test_repeater_chained_with_filter(self):
        """
        Chaining a Filter with a FilterRepeater causes the chained
        Filter to operate on the entire collection.
        """
        # This chain will apply NotEmpty to every item in the
        # collection, and then apply MaxLength to the collection as a
        # whole.
        self.filter_type =\
            lambda: f.FilterRepeater(f.NotEmpty) | f.MaxLength(2)

        # The collection has a length of 2, so it passes the MaxLength
        # filter.
        self.assertFilterPasses(['foo', 'bar'])

        # The collection has a length of 3, so it fails the MaxLength
        # filter.
        self.assertFilterErrors(
            ['a', 'b', 'c'],
            [f.MaxLength.CODE_TOO_LONG],
        )

    # noinspection SpellCheckingInspection
    def test_repeaterception(self):
        """
        FilterRepeaters can contain other FilterRepeaters.
        """
        self.filter_type = lambda: (
            # Apply the following filters to each item in the incoming
            # value:
            f.FilterRepeater(

                    # 1. It must be a list.
                    f.Type(list)

                    # 2. Apply the Int filter to each of its items.
                |   f.FilterRepeater(f.Int)

                    # 3. It must have a length <= 3.
                |   f.MaxLength(3)
            )
        )

        self.assertFilterPasses(
            #
            # Note that the INCOMING VALUE ITSELF does not have to be a
            # list, nor does it have to have a max length <= 3.
            #
            # These Filters are applied to the items INSIDE THE
            # INCOMING VALUE (because of the outer FilterRepeater).
            #
            {
                'foo':      ['1', '2', '3'],
                'bar':      [-20, 20],
                'baz':      ['486'],
                'luhrmann': [None, None, None],
            },

            {
                'foo':      [1, 2, 3],
                'bar':      [-20, 20],
                'baz':      [486],
                'luhrmann': [None, None, None],
            },
        )

        # The 1st item in this value is not a list, so it fails.
        self.assertFilterErrors(
            [
                [42],
                {'arch': 486},
            ],

            {
                '1': [f.Type.CODE_WRONG_TYPE],
            },

            expected_value = [[42], None],
        )

        # The 1st item in this value contains invalid ints.
        self.assertFilterErrors(
            [
                [42],
                ['NaN', 3.14, 'FOO'],
            ],

            {
                #
                # The error keys are the dotted paths to the invalid
                # values (in this case, they are numeric because we
                # are working with lists).
                #
                # This way, we don't have to deal with nested dicts
                # when processing error codes.
                #
                '1.0':  [f.Decimal.CODE_NON_FINITE],
                '1.1':  [f.Int.CODE_DECIMAL],
                '1.2':  [f.Decimal.CODE_INVALID],
            },

            expected_value = [[42], [None, None, None]],
        )

        # The 1st item in this value is too long.
        self.assertFilterErrors(
            [
                [42],
                [1, 2, 3, 4]
            ],

            {
                '1': [f.MaxLength.CODE_TOO_LONG],
            },

            expected_value = [[42], None]
        )


class FilterMapperTestCase(BaseFilterTestCase):
    def test_pass_none(self):
        """
        For consistency with all the other Filter classes, `None` is
        considered a valid value to pass to a FilterMapper, even
        though it is not iterable.
        """
        self.filter_type = lambda: f.FilterMapper({'id': f.Int})

        self.assertFilterPasses(None)

    def test_pass_mapping(self):
        """
        A FilterRepeater is applied to a dict containing valid values.
        """
        self.filter_type = lambda: f.FilterMapper({
            'id':       f.Required | f.Int | f.Min(1),
            'subject':  f.NotEmpty | f.MaxLength(16),
        })

        filter_ = self._filter({
            'id':       '42',
            'subject':  'Hello, world!',
        })

        self.assertFilterPasses(
            filter_,

            {
                'id':       42,
                'subject':  'Hello, world!',
            },
        )

        # The result is a dict, to match the type of the filter map.
        self.assertIs(type(filter_.cleaned_data), dict)

    def test_pass_ordered_mapping(self):
        """
        Configuring the FilterRepeater to return an OrderedDict.
        """
        # Note that we pass an OrderedDict to the filter initializer.
        self.filter_type = lambda: f.FilterMapper(OrderedDict((
            ('subject', f.NotEmpty | f.MaxLength(16)),
            ('id', f.Required | f.Int | f.Min(1)),
        )))

        filter_ = self._filter({
            'id':       '42',
            'subject':  'Hello, world!',
        })

        self.assertFilterPasses(
            filter_,

            OrderedDict((
                ('subject', 'Hello, world!'),
                ('id', 42),
            )),
        )

        # The result is an OrderedDict, to match the type of the filter
        # map.
        self.assertIs(type(filter_.cleaned_data), OrderedDict)

    def test_fail_mapping(self):
        """
        A FilterRepeater is applied to a dict containing invalid
        values.
        """
        self.filter_type = lambda: f.FilterMapper({
            'id':       f.Required | f.Int | f.Min(1),
            'subject':  f.NotEmpty | f.MaxLength(16),
        })

        self.assertFilterErrors(
            {
                'id':       None,
                'subject':  'Antidisestablishmentarianism',
            },

            {
                'id':       [f.Required.CODE_EMPTY],
                'subject':  [f.MaxLength.CODE_TOO_LONG],
            },

            expected_value = {
                'id':       None,
                'subject':  None,
            }
        )

    def test_extra_keys_allowed(self):
        """
        By default, FilterMappers passthru extra keys.
        """
        self.filter_type = lambda: f.FilterMapper({
            'id':       f.Required | f.Int | f.Min(1),
            'subject':  f.NotEmpty | f.MaxLength(16),
        })

        self.assertFilterPasses(
            {
                'id':       '42',
                'subject':  'Hello, world!',
                'extra':    'ignored',
            },

            {
                'id':       42,
                'subject':  'Hello, world!',
                'extra':    'ignored',
            }
        )

    def test_extra_keys_ordered(self):
        """
        When the filter map is an OrderedDict, extra keys are
        alphabetized.
        """
        # Note that we pass an OrderedDict to the filter initializer.
        self.filter_type = lambda: f.FilterMapper(OrderedDict((
            ('subject', f.NotEmpty | f.MaxLength(16)),
            ('id', f.Required | f.Int | f.Min(1)),
        )))

        filter_ = self._filter({
            'id':       '42',
            'subject':  'Hello, world!',
            'cat':      'felix',
            'bird':     'phoenix',
            'fox':      'fennecs',
        })

        self.assertFilterPasses(
            filter_,

            OrderedDict((
                # The filtered keys are always listed first.
                ('subject', 'Hello, world!'),
                ('id', 42),

                # Extra keys are listed afterward, in alphabetical
                # order.
                ('bird', 'phoenix'),
                ('cat', 'felix'),
                ('fox', 'fennecs'),
            )),
        )

    def test_extra_keys_disallowed(self):
        """
        FilterMappers can be configured to treat any extra key as an
        invalid value.
        """
        self.filter_type = lambda: f.FilterMapper(
            {
                'id':       f.Required | f.Int | f.Min(1),
                'subject':  f.NotEmpty | f.MaxLength(16),
            },

            # Treat all extra keys as invalid values.s
            allow_extra_keys = False,
        )

        self.assertFilterErrors(
            {
                'id':       '42',
                'subject':  'Hello, world!',
                'extra':    'ignored',
            },

            {
                'extra': [f.FilterMapper.CODE_EXTRA_KEY],
            },

            # The valid fields were still included in the return value,
            # but the invalid field was removed.
            expected_value = {
                'id':       42,
                'subject':  'Hello, world!',
            }
        )

    def test_extra_keys_specified(self):
        """
        FilterMappers can be configured only to allow certain extra
        keys.
        """
        self.filter_type = lambda: f.FilterMapper(
            {
                'id':       f.Required | f.Int | f.Min(1),
                'subject':  f.NotEmpty | f.MaxLength(16),
            },

            allow_extra_keys = {'message', 'extra'},
        )

        # As long as the extra keys are in the FilterMapper's
        # ``allow_extra_keys`` setting, everything is fine.
        self.assertFilterPasses(
            {
                'id':       '42',
                'subject':  'Hello, world!',
                'extra':    'ignored',
            },

            {
                'id':       42,
                'subject':  'Hello, world!',
                'extra':    'ignored',
            },
        )

        # But, add a key that isn't in ``allow_extra_keys``, and you've
        # got a problem.
        self.assertFilterErrors(
            {
                'id':           '42',
                'subject':      'Hello, world!',
                'attachment':   {
                    'type':         'image/jpeg',
                    'data':         '...',
                },
            },

            {
                'attachment': [f.FilterMapper.CODE_EXTRA_KEY],
            },

            expected_value = {
                'id':       42,
                'subject':  'Hello, world!',
            }
        )

    def test_missing_keys_allowed(self):
        """
        By default, FilterMappers treat missing keys as `None`.
        """
        self.filter_type = lambda: f.FilterMapper({
            'id':       f.Required | f.Int | f.Min(1),
            'subject':  f.NotEmpty | f.MaxLength(16),
        })

        # 'subject' allows null values, so no errors are generated.
        self.assertFilterPasses(
            {
                'id':   '42',
            },

            {
                'id':       42,
                'subject':  None,
            },
        )

        # However, 'id' has Required in its FilterChain, so a missing
        # 'id' is still an error.
        self.assertFilterErrors(
            {
                'subject': 'Hello, world!',
            },

            {
                'id': [f.Required.CODE_EMPTY],
            },

            expected_value = {
                'id':       None,
                'subject':  'Hello, world!',
            },
        )

    def test_missing_keys_disallowed(self):
        """
        FilterMappers can be configured to treat missing keys as
        invalid values.
        """
        self.filter_type = lambda: f.FilterMapper(
            {
                'id':       f.Required | f.Int | f.Min(1),
                'subject':  f.NotEmpty | f.MaxLength(16),
            },

            # Treat missing keys as invalid values.
            allow_missing_keys = False,
        )

        self.assertFilterErrors(
            {},

            {
                'id':       [f.FilterMapper.CODE_MISSING_KEY],
                'subject':  [f.FilterMapper.CODE_MISSING_KEY],
            },

            expected_value = {
                'id':       None,
                'subject':  None,
            },
        )

    def test_missing_keys_specified(self):
        """
        FilterMappers can be configured to allow some missing keys but
        not others.
        """
        self.filter_type = lambda: f.FilterMapper(
            {
                'id':       f.Required | f.Int | f.Min(1),
                'subject':  f.NotEmpty | f.MaxLength(16),
            },

            allow_missing_keys = {'subject'},
        )

        # The FilterMapper is configured to treat missing 'subject' as
        # if it were set to `None`.
        self.assertFilterPasses(
            {
                'id': '42'
            },

            {
                'id':       42,
                'subject':  None,
            },
        )

        # However, 'id' is still required.
        self.assertFilterErrors(
            {
                'subject':  'Hello, world!',
            },

            {
                'id': [f.FilterMapper.CODE_MISSING_KEY],
            },

            expected_value = {
                'id':       None,
                'subject':  'Hello, world!',
            }
        )

    def test_passthru_key(self):
        """
        If you want to make a key required but do not want to run any
        Filters on it, set its FilterChain to `None`.
        """
        self.filter_type = lambda: f.FilterMapper(
            {
                'id':       f.Required | f.Int | f.Min(1),
                'subject':  None,
            },

            # If you configure a FilterMapper with passthru keys(s),
            # you generally also want to disallow missing keys.
            allow_missing_keys = False,
        )

        self.assertFilterPasses(
            {
                'id':       '42',
                'subject':  'Hello, world!',
            },

            {
                'id':       42,
                'subject':  'Hello, world!',
            },
        )

        self.assertFilterPasses(
            {
                'id':       '42',
                'subject':  None,
            },

            {
                'id':       42,
                'subject':  None,
            },
        )

        self.assertFilterErrors(
            {
                'id': '42',
            },

            {
                'subject': [f.FilterMapper.CODE_MISSING_KEY],
            },

            expected_value = {
                'id':       42,
                'subject':  None,
            },
        )

    def test_fail_non_mapping(self):
        """The incoming value is not a mapping."""
        self.filter_type = lambda: f.FilterMapper({
            'id':       f.Required | f.Int | f.Min(1),
            'subject':  f.NotEmpty | f.MaxLength(16),
        })

        self.assertFilterErrors(
            # Nope; it's gotta be an explicit mapping.
            (('id', '42'), ('subject', 'Hello, world!')),
            [f.Type.CODE_WRONG_TYPE],
        )

    def test_mapper_chained_with_mapper(self):
        """
        Chaining two FilterMappers together has basically the same
        effect as combining their Filters.

        Generally, combining two FilterMappers into a single instance
        is much easier to read/maintain than chaining them, but in
        a few cases it may be unavoidable (for example, if you need
        each FilterMapper to handle extra and/or missing keys
        differently).
        """
        fm1 = f.FilterMapper(
            {
                'id': f.Int | f.Min(1),
            },

            allow_missing_keys  = True,
            allow_extra_keys    = True,
        )

        fm2 = f.FilterMapper(
            {
                'id':       f.Required | f.Max(256),
                'subject':  f.NotEmpty | f.MaxLength(16),
            },

            allow_missing_keys  = False,
            allow_extra_keys    = False,
        )

        self.filter_type = lambda: fm1 | fm2

        self.assertFilterPasses(
            {
                'id':       '42',
                'subject':  'Hello, world!',
            },

            {
                'id':       42,
                'subject':  'Hello, world!',
            },
        )

        self.assertFilterErrors(
            {},

            {
                # ``fm1`` allows missing keys, so it sets 'id' to
                # ``None``.
                # However, ``fm2`` does not allow ``None`` for 'id'
                # (because of the ``Required`` filter).
                'id':       [f.Required.CODE_EMPTY],

                # `fm1` does not care about `subject`, but `fm2`
                # expects it to be there.
                'subject':  [f.FilterMapper.CODE_MISSING_KEY],
            },

            expected_value = {
                'id':       None,
                'subject':  None,
            },
        )

    def test_filter_mapper_chained_with_filter(self):
        """
        Chaining a Filter with a FilterMapper causes the chained Filter
        to operate on the entire mapping.
        """
        fm = f.FilterMapper({
            'id':       f.Required | f.Int | f.Min(1),
            'subject':  f.NotEmpty | f.MaxLength(16),
        })

        self.filter_type = lambda: fm | f.MaxLength(3)

        self.assertFilterPasses(
            {
                'id':       '42',
                'subject':  'Hello, world!',
                'extra':    'ignored',
            },

            {
                'id':       42,
                'subject':  'Hello, world!',
                'extra':    'ignored',
            },
        )

        self.assertFilterErrors(
            {
                'id':           '42',
                'subject':      'Hello, world!',
                'extra':        'ignored',
                'attachment':   None,
            },

            # The incoming value has 4 items, which fails the MaxLength
            # filter.
            [f.MaxLength.CODE_TOO_LONG],
        )

    # noinspection SpellCheckingInspection
    def test_mapperception(self):
        """
        Want to filter dicts that contain other dicts?
        We need to go deeper.
        """
        self.filter_type = lambda: f.FilterMapper(
            {
                'id':           f.Required | f.Int | f.Min(1),
                'subject':      f.NotEmpty | f.MaxLength(16),
                'attachment':   f.FilterMapper(
                    {
                        'type':
                                f.Required
                            |   f.Choice(choices={'image/jpeg', 'image/png'}),

                        'data': f.Required | f.Base64Decode,
                    },

                    allow_extra_keys    = False,
                    allow_missing_keys  = False,
                )
            },

            allow_extra_keys    = False,
            allow_missing_keys  = False,
        )

        # Valid mapping is valid.
        self.assertFilterPasses(
            {
                'id':      '42',
                'subject': 'Hello, world!',

                'attachment': {
                    'type': 'image/jpeg',

                    'data':
                        b'R0lGODlhDwAPAKECAAAAzMzM/////wAAACwAAAAAD'
                        b'wAPAAACIISPeQHsrZ5ModrLlN48CXF8m2iQ3YmmKq'
                        b'VlRtW4MLwWACH+EVRIRSBDQUtFIElTIEEgTElFOw==',
                },
            },

            {
                'id':       42,
                'subject':  'Hello, world!',

                'attachment': {
                    'type': 'image/jpeg',

                    'data':
                        b'GIF89a\x0f\x00\x0f\x00\xa1\x02\x00\x00\x00'
                        b'\xcc\xcc\xcc\xff\xff\xff\xff\x00\x00\x00,\x00'
                        b'\x00\x00\x00\x0f\x00\x0f\x00\x00\x02 \x84\x8f'
                        b'y\x01\xec\xad\x9eL\xa1\xda\xcb\x94\xde<\tq|'
                        b'\x9bh\x90\xdd\x89\xa6*\xa5eF\xd5\xb80\xbc\x16'
                        b'\x00!\xfe\x11THE CAKE IS A LIE;'
                },
            },
        )

        # Invalid mapping... not so much.
        self.assertFilterErrors(
            {
                'id': 'NaN',

                'attachment': {
                    'type': 'foo',
                    'data': False,
                },
            },

            {
                # The error keys are the dotted paths to the invalid
                # values.
                # This way, we don't have to deal with nested dicts
                # when processing error codes.
                'id':               [f.Decimal.CODE_NON_FINITE],
                'subject':          [f.FilterMapper.CODE_MISSING_KEY],
                'attachment.type':  [f.Choice.CODE_INVALID],
                'attachment.data':  [f.Type.CODE_WRONG_TYPE],
            },

            # The resulting value has the expected structure, but it's
            # a ghost town.
            expected_value = {
                'id':       None,
                'subject':  None,

                'attachment': {
                    'type': None,
                    'data': None,
                },
            },
        )



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from datetime import datetime
from unittest import TestCase

from pytz import utc

import filters as f
from filters.macros import filter_macro, FilterMacroType


class FilterMacroTestCase(TestCase):
    """
    Unit tests for :py:func:`filter_macro`.
    """
    def test_decorator(self):
        """
        A common use case for Filter macros is to use them as
        decorators for functions.
        """
        @filter_macro
        def MyFilter():
            return f.Unicode | f.Strip | f.MinLength(12)

        # You can use :py:class:`FilterMacroType` to detect a filter
        # macro.
        self.assertTrue(issubclass(MyFilter, FilterMacroType))

        # As you would expect, invoking the macro returns a
        # FilterChain.
        the_filter = MyFilter()

        self.assertNotIsInstance(the_filter, FilterMacroType)
        self.assertIsInstance(the_filter, f.FilterChain)

        self.assertEqual(
            the_filter.apply('  Hello, world!  '),
            'Hello, world!',
        )

        with self.assertRaises(f.FilterError):
            the_filter.apply('Hi there!')

    def test_chain(self):
        """
        You can chain Filter macros with other Filters, the same as you
        would with any other Filter.
        """
        @filter_macro
        def MyFilter():
            return f.Unicode | f.Strip | f.MinLength(12)

        # Note that you don't have to invoke the macro to include it in
        # a chain.
        # If you don't believe me, try removing the decorator from
        # ``MyFilter``, and watch this test explode.
        filter_chain = MyFilter | f.Split(r'\s+')

        self.assertEqual(
            filter_chain.apply('Hello, world!'),
            ['Hello,', 'world!'],
        )

        with self.assertRaises(f.FilterError):
            filter_chain.apply('Hi there!')

    def test_chain_macros(self):
        """
        Yup, you can chain Filter macros together, too.
        """
        @filter_macro
        def Filter1():
            return f.Unicode | f.Strip

        @filter_macro
        def Filter2():
            return f.Split(r'\s+') | f.MinLength(2)

        filter_chain = Filter1 | Filter2

        self.assertEqual(
            filter_chain.apply('  Hello, world!  '),
            ['Hello,', 'world!'],
        )

        with self.assertRaises(f.FilterError):
            filter_chain.apply('whazzup!')

    def test_decorator_optional_parameters(self):
        """
        A filter macro may accept optional parameters.
        """
        @filter_macro
        def MyFilter(min_length=12):
            return f.Unicode | f.MinLength(min_length)

        # `MyFilter` is configured to require 12 chars by default.
        filter_chain = f.Required | MyFilter

        self.assertEqual(
            filter_chain.apply('Hello, world!'),
            'Hello, world!',
        )

        with self.assertRaises(f.FilterError):
            filter_chain.apply('Hi there!')

    def test_partial(self):
        """
        You can use Filter macros to create partials from other Filter
        types.
        """
        MyDatetime = filter_macro(f.Datetime, timezone=12)

        self.assertEqual(
            MyDatetime().apply('2015-10-13 15:22:18'),

            # By default, MyDatetime assumes a timezone of UTC+12...
            datetime(2015, 10, 13, 3, 22, 18, tzinfo=utc),
        )

        self.assertEqual(
            # ... however, we can override it.
            MyDatetime(timezone=6).apply('2015-10-13 15:22:18'),

            datetime(2015, 10, 13, 9, 22, 18, tzinfo=utc),
        )



# coding=utf-8
from __future__ import absolute_import, unicode_literals

from abc import ABCMeta
from functools import partial, WRAPPER_ASSIGNMENTS
from six import with_metaclass

from filters.base import BaseFilter, FilterMeta

__all__ = [
    'FilterMacroType',
    'filter_macro',
]


class FilterMacroType(with_metaclass(ABCMeta, BaseFilter)):
    """
    Base type for filter macros.  Doesn't do anything on its own, but
    it is useful for identifying filter macros when paired with an
    ``issubclass`` check.

    Important:  Use ``issubclass``, not ``isinstance``!

    .. code:: python

       @filter_macro
       def MyMacro():
         return f.NoOp

       # ``MyMacro`` is now a *subclass* of ``FilterMacroType``!
       assert issubclass(MyMacro, FilterMacroType)

       # It is *not* an *instance* of ``FilterMacroType``!
       assert not isinstance(MyMacro, FilterMacroType)
    """
    pass


def filter_macro(func, *args, **kwargs):
    """
    Promotes a function that returns a filter into its own filter type.

    Example::

        @filter_macro
        def String():
            return Unicode | Strip | NotEmpty

        # You can now use `String` anywhere you would use a regular Filter:
        (String | Split(':')).apply('...')

    You can also use ``filter_macro`` to create partials, allowing you to
    preset one or more initialization arguments::

        Minor = filter_macro(Max, max_value=18, inclusive=False)
        Minor(inclusive=True).apply(18)
    """
    filter_partial = partial(func, *args, **kwargs)

    class FilterMacroMeta(FilterMeta):
        @staticmethod
        def __new__(mcs, name, bases, attrs):
            # This is as close as we can get to running
            # ``update_wrapper`` on a type.
            for attr in WRAPPER_ASSIGNMENTS:
                if hasattr(func, attr):
                    attrs[attr] = getattr(func, attr)

            # Note that we ignore the ``name`` argument, passing in
            # ``func.__name__`` instead.
            return super(FilterMacroMeta, mcs)\
                .__new__(mcs, func.__name__, bases, attrs)

        def __call__(cls, *runtime_args, **runtime_kwargs):
            return filter_partial(*runtime_args, **runtime_kwargs)

    class FilterMacro(with_metaclass(FilterMacroMeta, FilterMacroType)):
        # This method will probably never get called due to overloaded
        # ``__call__`` in the metaclass, but just in case, we'll include
        # it because it is an abstract method in `BaseFilter`.
        def _apply(self, value):
            # noinspection PyProtectedMember
            return self.__class__()._apply(value)

    return FilterMacro



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from typing import Optional, Text

import filters as f
from filters.macros import filter_macro


# noinspection PyClassHasNoInit
class TestFilterAlpha(f.BaseFilter):
    """
    A filter that can be used for testing.
    """
    def _apply(self, value):
        return value


# noinspection PyClassHasNoInit
class TestFilterBravo(f.BaseFilter):
    """
    A filter that will can be used for testing.
    """
    def __init__(self, name=None):
        # type: (Optional[Text]) -> None
        super(TestFilterBravo, self).__init__()

        self.name = name

    def _apply(self, value):
        return value


@filter_macro
def TestFilterCharlie():
    """
    A filter macro that can be used for testing.
    """
    return f.NoOp



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from decimal import Decimal
from uuid import UUID
from xml.etree.ElementTree import Element

import re
# noinspection PyCompatibility
import regex
from collections import OrderedDict

from six import itervalues

import filters as f
from filters.test import BaseFilterTestCase

from test.simple_test import Unicody, Bytesy


class Base64DecodeTestCase(BaseFilterTestCase):
    filter_type = f.Base64Decode

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use ``Required | Base64Decode`` if you want to reject null
        values.
        """
        self.assertFilterPasses(None)

    def test_pass_valid(self):
        """
        The incoming value is Base64-encoded.
        """
        # noinspection SpellCheckingInspection
        self.assertFilterPasses(b'SGVsbG8sIHdvcmxkIQ==', b'Hello, world!')

    def test_pass_url_safe(self):
        """
        The incoming value is Base64-encoded using a URL-safe variant.

        This actually doesn't happen that often, as most human-readable
        strings tend not to contain any URL-unsafe characters when
        they are base64-encoded.
        """
        self.assertFilterPasses(
            b'--___w==',
            b'\xfb\xef\xff\xff',
        )

    def test_fail_mixed_dialects(self):
        """
        The incoming value contains both URL-safe and URL-unsafe
        characters.
        """
        self.assertFilterErrors(
            b'+-_/_w==',
            [f.Base64Decode.CODE_INVALID]
        )

    def test_pass_whitespace(self):
        """
        The incoming value includes whitespace characters.

        Technically, whitespace chars are not part of the Base64
        alphabet.  But, virtually every implementation includes
        support for whitespace, so we will, too.
        """
        self.assertFilterPasses(
            # Tab chars are especially weird, but eh, why not..
            b'SGV sbG 8sI\tHdv\ncmx\rkIQ\r\n',
            b'Hello, world!',
        )

    def test_pass_padding_missing(self):
        """
        The incoming value is Base64-encoded, but it has the wrong
        length.

        Base64 works by splitting up the string into chunks of 3 bytes
        (24 bits) each, then dividing each chunk into 4 smaller
        chunks of 6 bits each.  If the string's length is not
        divisible by 3, then the last chunk will have too few
        bytes, so we have to pad it out.

        References:
          - https://en.wikipedia.org/wiki/Base64#Padding
        """
        # noinspection SpellCheckingInspection
        self.assertFilterPasses(b'SGVsbG8sIHdvcmxkIQ', b'Hello, world!')

    def test_pass_padding_excessive(self):
        """
        The incoming value is Base64-encoded, but for some reason it
        has too much padding.
        """
        # noinspection SpellCheckingInspection
        self.assertFilterPasses(b'SGVsbG8sIHdvcmxkIQ=====', b'Hello, world!')

    def test_fail_invalid(self):
        """
        The incoming value is not Base64-encoded.
        """
        self.assertFilterErrors(
            # Interestingly, Python's ``base64`` functions will attempt to
            # decode this string anyway, by ignoring the invalid
            # characters.
            # https://docs.python.org/2/library/base64.html#base64.b64decode
            b'$Hello, world!===$',
            [f.Base64Decode.CODE_INVALID],
        )

    def test_fail_string(self):
        """
        To ensure consistent behavior between Python 2 and Python 3,
        character strings are not allowed; only binary strings can be
        decoded.
        """
        # noinspection SpellCheckingInspection
        self.assertFilterErrors(
            'SGVsbG8sIHdvcmxkIQ==',
            [f.Type.CODE_WRONG_TYPE],
        )

    def test_fail_wrong_type(self):
        """
        The incoming value is not a string.
        """
        # noinspection SpellCheckingInspection
        self.assertFilterErrors(
            [b'kB1ReXKFSE5xgu0sODTVrJWg4eYDz32iRLm+fATfsBQ='],
            [f.Type.CODE_WRONG_TYPE],
        )


class ByteStringTestCase(BaseFilterTestCase):
    filter_type = f.ByteString

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use ``Required | ByteString`` if you want to reject null values.
        """
        self.assertFilterPasses(None)

    def test_pass_unicode(self):
        """
        The incoming value is a unicode.
        """
        self.assertFilterPasses(
            'Iñtërnâtiônàlizætiøn',

            # 'Iñtërnâtiônàlizætiøn' encoded as bytes using utf-8:
            b'I\xc3\xb1t\xc3\xabrn\xc3\xa2ti\xc3'
            b'\xb4n\xc3\xa0liz\xc3\xa6ti\xc3\xb8n',
        )

    def test_pass_bytes_utf8(self):
        """
        The incoming value is a byte string, already encoded as UTF-8.
        """
        self.assertFilterPasses(
            b'I\xc3\xb1t\xc3\xabrn\xc3\xa2ti\xc3'
            b'\xb4n\xc3\xa0liz\xc3\xa6ti\xc3\xb8n'
        )

    def test_fail_bytes_non_utf8(self):
        """
        The incoming value is a byte string, but encoded using a
        different codec.
        """
        # 'Iñtërnâtiônàlizætiøn' encoded as bytes using ISO-8859-1:
        incoming = b'I\xf1t\xebrn\xe2ti\xf4n\xe0liz\xe6ti\xf8n'

        self.assertFilterErrors(
            incoming,
            [f.ByteString.CODE_DECODE_ERROR],
        )

        # In order for this to work, we have to tell the Filter what
        # encoding to use:
        self.assertFilterPasses(
            self._filter(incoming, encoding='iso-8859-1'),

            # The result is re-encoded using UTF-8.
            b'I\xc3\xb1t\xc3\xabrn\xc3\xa2ti\xc3'
            b'\xb4n\xc3\xa0liz\xc3\xa6ti\xc3\xb8n',
        )

    def test_pass_string_like_object(self):
        """
        The incoming value is an object that can be cast as a unicode.
        """
        self.assertFilterPasses(
            Unicody('༼ つ ◕_◕ ༽つ'),

            # Stoned Kirby?  Jigglypuff in dance mode?
            # I have no idea what this is.
            b'\xe0\xbc\xbc \xe3\x81\xa4 \xe2\x97\x95_'
            b'\xe2\x97\x95 \xe0\xbc\xbd\xe3\x81\xa4',
        )

    def test_pass_bytes_like_object(self):
        """
        The incoming value is an object that can be cast as a byte
        string.
        """
        value = (
            # Person
            b'(\xe2\x95\xaf\xc2\xb0\xe2\x96\xa1\xc2\xb0)'
            # Particle Effects
            b'\xe2\x95\xaf\xef\xb8\xb5 '
            # Table
            b'\xe2\x94\xbb\xe2\x94\x81\xe2\x94\xbb'
        )

        self.assertFilterPasses(Bytesy(value), value)

    def test_pass_boolean(self):
        """
        The incoming value is a boolean (treated as an int).
        """
        self.assertFilterPasses(True, b'1')

    def test_pass_decimal_with_scientific_notation(self):
        """
        The incoming value is a Decimal that was parsed from scientific
        notation.
        """
        # Note that ``bytes(Decimal('2.8E6')`` yields b'2.8E+6', which
        # is not what we want!
        self.assertFilterPasses(
            Decimal('2.8E6'),
            b'2800000',
        )

    def test_pass_xml_element(self):
        """
        The incoming value is an ElementTree XML Element.
        """
        self.assertFilterPasses(
            Element('foobar'),
            b'<foobar />',
        )

    def test_unicode_normalization_off_by_default(self):
        """
        By default, the Filter does not apply normalization before
        encoding.

        References:
          - https://en.wikipedia.org/wiki/Unicode_equivalence
          - http://stackoverflow.com/q/16467479
        """
        self.assertFilterPasses(
            #   U+0065 LATIN SMALL LETTER E
            # + U+0301 COMBINING ACUTE ACCENT
            # (2 code points)
            'Ame\u0301lie',

            # Result is the same string, encoded using UTF-8.
            b'Ame\xcc\x81lie',
        )

    def test_unicode_normalization_forced(self):
        """
        You can force the Filter to apply normalization before encoding.

        References:
          - https://en.wikipedia.org/wiki/Unicode_equivalence
          - http://stackoverflow.com/q/16467479
        """
        self.assertFilterPasses(
            self._filter(
                # Same decomposed sequence from previous test...
                'Ame\u0301lie',

                # ... but this time we tell the Filter to normalize the
                # value before encoding it.
                normalize = True,
            ),

            # U+00E9 LATIN SMALL LETTER E WITH ACUTE
            # (1 code point, encoded as bytes)
            b'Am\xc3\xa9lie',
        )

    def test_remove_non_printables_off_by_default(self):
        """
        By default, the Filter does not remove non-printable
        characters.
        """
        self.assertFilterPasses(
            # \u0000-\u001f are ASCII control characters.
            # \uffff is a Unicode control character.
            '\u0010Hell\u0000o,\u001f wor\uffffld!',

            b'\x10Hell\x00o,\x1f wor\xef\xbf\xbfld!',
        )

    def test_remove_non_printables_forced(self):
        """
        You can force the Filter to remove non-printable characters
        before encoding.
        """
        self.assertFilterPasses(
            self._filter(
                '\u0010Hell\u0000o,\u001f wor\uffffld!',
                normalize = True,
            ),
            b'Hello, world!',
        )

    def test_newline_normalization_off_by_default(self):
        """
        By default, the Filter does not normalize line endings.
        """
        self.assertFilterPasses(
            'unix\n - windows\r\n - weird\r\r\n',
            b'unix\n - windows\r\n - weird\r\r\n',
        )

    def test_newline_normalization_forced(self):
        """
        You can force the Filter to normalize line endings.
        """
        self.assertFilterPasses(
            self._filter('unix\n - windows\r\n - weird\r\r\n', normalize=True),
            b'unix\n - windows\n - weird\n\n',
        )


class CaseFoldTestCase(BaseFilterTestCase):
    filter_type = f.CaseFold

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use ``Required | CaseFold`` if you want to reject null values.
        """
        self.assertFilterPasses(None)

    def test_pass_ascii(self):
        """
        The incoming value is ASCII.
        """
        self.assertFilterPasses('FOO bar BAZ', 'foo bar baz')

    def test_pass_unicode(self):
        """
        The incoming value is not ASCII.
        """
        # For some reason, the internet really loves to use ß to test
        # case folding functionality.
        self.assertFilterPasses('Weißkopfseeadler', 'weisskopfseeadler')

        # Note that case-folded does not necessarily mean ASCII-
        # compatible!
        # noinspection SpellCheckingInspection
        self.assertFilterPasses('İstanbul', 'i\u0307stanbul')

    def test_pass_unfoldable(self):
        """
        There are some Unicode characters that look foldable but
        actually aren't.

        Spotify learned this the hard way.

        References:
          - https://labs.spotify.com/2013/06/18/creative-usernames/
        """
        self.assertFilterPasses(u'\u1d2e\u1d35\u1d33\u1d2e\u1d35\u1d3f\u1d30')

    def test_fail_bytes(self):
        """
        To ensure consistent behavior in Python 2 and Python 3, byte
        strings are not allowed.
        """
        self.assertFilterErrors(
            b"look im already folded anyway",
            [f.Type.CODE_WRONG_TYPE],
        )

    def test_fail_wrong_type(self):
        """
        The incoming value is not a string.
        """
        self.assertFilterErrors(
            ['Weißkopfseeadler', 'İstanbul'],
            [f.Type.CODE_WRONG_TYPE],
        )


class IpAddressTestCase(BaseFilterTestCase):
    filter_type = f.IpAddress

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use ``Required | IpAddress`` if you want to reject null values.
        """
        self.assertFilterPasses(None)

    def test_ipv4_success_happy_path(self):
        """
        The incoming value is a valid IPv4 address.
        """
        self.assertFilterPasses('127.0.0.1')

    def test_ipv4_error_invalid(self):
        """
        The incoming value is not a valid IPv4 address.
        """
        self.assertFilterErrors('127.0.0.1/32', [f.IpAddress.CODE_INVALID])
        self.assertFilterErrors('256.0.0.1', [f.IpAddress.CODE_INVALID])
        self.assertFilterErrors('-1.0.0.1', [f.IpAddress.CODE_INVALID])

    def test_ipv4_error_too_short(self):
        """
        Technically, an IPv4 address can contain less than 4 octets,
        but this Filter always expects exactly 4.
        """
        self.assertFilterErrors('127.1', [f.IpAddress.CODE_INVALID])

    def test_ipv4_error_too_long(self):
        """
        The incoming value looks like an IPv4 address, except it
        contains too many octets.
        """
        self.assertFilterErrors('127.0.0.1.32', [f.IpAddress.CODE_INVALID])

    def test_ipv4_error_ipv6(self):
        """
        By default, this Filter does not accept IPv6 addresses.
        """
        self.assertFilterErrors(
            '2001:0db8:85a3:0000:0000:8a2e:0370:7334',
            [f.IpAddress.CODE_INVALID],
        )

    def test_ipv6_success_happy_path(self):
        """
        The incoming value is a valid IPv6 address.
        """
        self.assertFilterPasses(
            self._filter(
                '2001:0db8:85a3:0000:0000:8a2e:0370:7334',

                # You must explicitly configure the filter to accept
                # IPv6 addresses.
                ipv4    = False,
                ipv6    = True,
            ),

            # Note that the resulting value is automatically
            # abbreviated, if possible.
            # https://en.wikipedia.org/wiki/IPv6_address#Presentation
            '2001:db8:85a3::8a2e:370:7334',
        )

    def test_ipv6_success_case_insensitive(self):
        """
        The incoming value uses mixed case for hex characters.
        """
        self.assertFilterPasses(
            self._filter(
                '2001:0DB8:85A3:0000:0000:8a2e:0370:7334',
                ipv4 = False,
                ipv6 = True,
            ),

            '2001:db8:85a3::8a2e:370:7334',
        )

    def test_ipv6_success_truncated_zeroes(self):
        """
        IPv6 supports truncating leading zeroes.
        """
        self.assertFilterPasses(
            self._filter(
                '2001:db8:85a3:0:0:8a2e:370:7334',
                ipv4 = False,
                ipv6 = True,
            ),

            '2001:db8:85a3::8a2e:370:7334',
        )

    def test_ipv6_success_truncated_groups(self):
        """
        Empty groups (all zeroes) can be omitted entirely.
        """
        self.assertFilterPasses(
            self._filter(
                '2001:db8:85a3::8a2e:370:7334',
                ipv4 = False,
                ipv6 = True,
            ),
        )

    def test_ipv6_success_dotted_quad(self):
        """
        IPv6 supports "dotted quad" notation for IPv4 addresses that
        are mid-transition.
        """
        # noinspection SpellCheckingInspection
        self.assertFilterPasses(
            self._filter('::ffff:192.0.2.128', ipv4=False, ipv6=True),
        )

    def test_ipv6_error_invalid(self):
        """
        Invalid IPv6 address is invalid.
        """
        self.assertFilterErrors(
            self._filter('not even close', ipv4=False, ipv6=True),
            [f.IpAddress.CODE_INVALID],
        )

    def test_ipv6_error_too_long(self):
        """
        If the incoming value has too many groups to be IPv6, it is
        invalid.
        """
        self.assertFilterErrors(
            self._filter(
                # Oops; one group too many!
                '2001:0db8:85a3:0000:0000:8a2e:0370:7334:1234',
                ipv4 = False,
                ipv6 = True,
            ),
            [f.IpAddress.CODE_INVALID],
        )

    def test_ipv6_error_ipv4(self):
        """
        If the Filter is configured only to accept IPv6 addresses, IPv4
        addresses are invalid.
        """
        self.assertFilterErrors(
            self._filter('127.0.0.1', ipv4=False, ipv6=True),
            [f.IpAddress.CODE_INVALID],
        )

    def test_pass_allow_ipv4_and_ipv6(self):
        """
        You can configure the Filter to accept both IPv4 and IPv6
        addresses.
        """
        self.assertFilterPasses(
            self._filter('127.0.0.1', ipv4=True, ipv6=True),
        )

        self.assertFilterPasses(
            self._filter(
                '2001:0db8:85a3:0000:0000:8a2e:0370:7334',

                ipv4 = True,
                ipv6 = True,
            ),
            '2001:db8:85a3::8a2e:370:7334'
        )

    def test_fail_bytes(self):
        """
        To ensure consistent behavior in Python 2 and Python 3, byte
        strings are not allowed.
        """
        self.assertFilterErrors(b'127.0.0.1', [f.Type.CODE_WRONG_TYPE])

    def test_fail_wrong_type(self):
        """
        The incoming value is not a string.
        """
        self.assertFilterErrors(
            ['127.0.0.1', '192.168.1.1'],
            [f.Type.CODE_WRONG_TYPE],
        )


class JsonDecodeTestCase(BaseFilterTestCase):
    filter_type = f.JsonDecode

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use ``Required | JsonDecode`` if you want to reject null
        values.
        """
        self.assertFilterPasses(None)

    def test_pass_valid_json(self):
        """
        The incoming value is valid JSON.
        """
        self.assertFilterPasses(
            '{"foo": "bar", "baz": "luhrmann"}',

            # Technically, the return value is an OrderedDict, but to
            # keep things simple, we will compare them as dicts.
            {'foo': 'bar', 'baz': 'luhrmann'},
        )

    def test_fail_invalid_json(self):
        """
        The incoming value is not valid JSON.
        """
        self.assertFilterErrors(
            '{"almost_valid": true',
            [f.JsonDecode.CODE_INVALID],
        )

    def test_fail_empty_string(self):
        """
        The incoming value is an empty string.

        Consider using ``NotEmpty | Json`` so that users get more
        meaningful feedback for empty strings.
        """
        self.assertFilterErrors('', [f.JsonDecode.CODE_INVALID])

    def test_fail_bytes(self):
        """
        To ensure consistent behavior in Python 2 and Python 3, byte
        strings are not allowed.
        """
        self.assertFilterErrors(b'{"blends": false}', [f.Type.CODE_WRONG_TYPE])

    def test_fail_wrong_type(self):
        """
        The incoming value is not a string.
        """
        self.assertFilterErrors({'foo': 'bar'}, [f.Type.CODE_WRONG_TYPE])


class MaxBytesTestCase(BaseFilterTestCase):
    filter_type = f.MaxBytes

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use ``Required | MaxBytes`` if you want to reject null values.
        """
        self.assertFilterPasses(self._filter(None, max_bytes=65535))

    def test_pass_string_short(self):
        """
        The incoming value is a string that is short enough.
        """
        self.assertFilterPasses(
            self._filter('Γειάσου Κόσμε', max_bytes=25),

            # The filter always returns bytes.
            'Γειάσου Κόσμε'.encode('utf-8'),
        )

    def test_pass_string_short_with_prefix(self):
        """
        The filter is configured to apply a prefix to values that are
        too long, but the incoming value is a unicode that is short
        enough.
        """
        # If we were to apply the prefix to the incoming string, it
        # would be too long to fit, but the filter will only apply
        # the prefix if the value needs to be truncated.
        self.assertFilterPasses(
            self._filter('Γειάσου Κόσμε', max_bytes=25, prefix='σφάλμα:'),

            # The filter always returns bytes.
            'Γειάσου Κόσμε'.encode('utf-8'),
        )

    def test_fail_string_long(self):
        """
        The incoming value is a string that is too long.
        """
        self.assertFilterErrors(
            self._filter('Γειάσου Κόσμε', max_bytes=24),
            [f.MaxBytes.CODE_TOO_LONG],

            # Note that the resulting value is truncated to 23 bytes
            # instead of 24, so as not to orphan a multibyte
            # sequence.
            expected_value =
                b'\xce\x93\xce\xb5\xce\xb9\xce\xac\xcf\x83\xce\xbf'
                b'\xcf\x85 \xce\x9a\xcf\x8c\xcf\x83\xce\xbc',
        )

    def test_fail_string_long_with_prefix(self):
        """
        The incoming value is a string that is too long, and the
        filter is configured to apply a prefix before truncating.
        """
        self.assertFilterErrors(
            self._filter('Γειάσου Κόσμε', max_bytes=24, prefix='σφάλμα:'),
            [f.MaxBytes.CODE_TOO_LONG],

            # Note that the prefix reduces the number of bytes
            # available when truncating the value.
            expected_value =
                b'\xcf\x83\xcf\x86\xce\xac\xce\xbb\xce\xbc\xce\xb1:'
                b'\xce\x93\xce\xb5\xce\xb9\xce\xac\xcf\x83'
        )

    def test_fail_string_long_no_truncate(self):
        """
        The incoming value is a string that is too long, and the
        filter is configured not to truncate values.
        """
        self.assertFilterErrors(
            self._filter('Γειάσου Κόσμε', max_bytes=24, truncate=False),
            [f.MaxBytes.CODE_TOO_LONG],

            # When the filter is configured with `truncate=False`, it
            # returns `None` instead of truncating too-long values.
            expected_value = None,
        )

    def test_fail_string_tiny_max_bytes(self):
        """
        The filter is configured with a `max_bytes` so tiny that it is
        impossible to fit any multibyte sequences into a truncated
        string.
        """
        self.assertFilterErrors(
            self._filter('你好，世界！', max_bytes=2),
            [f.MaxBytes.CODE_TOO_LONG],

            # The filter returns an empty string, not `None`.
            expected_value = b'',
        )

    def test_pass_string_alt_encoding(self):
        """
        The filter is configured to use an encoding other than UTF-8,
        and the incoming value is a string that is short enough.
        """
        self.assertFilterPasses(
            self._filter(
                'Γειάσου Κόσμε',

                max_bytes   = 13,
                encoding    = 'iso-8859-7',
            ),

            # The resulting value is encoded using ISO-8859-7 (Latin-1
            # Greek).
            b'\xc3\xe5\xe9\xdc\xf3\xef\xf5 \xca\xfc\xf3\xec\xe5',
        )

    def test_fail_string_alt_encoding(self):
        """
        The filter is configured to use an encoding other that UTF-8,
        and the incoming value is a string that is too long.
        """
        self.assertFilterErrors(
            self._filter(
                'Γειάσου Κόσμε',

                max_bytes   = 13,
                encoding    = 'utf-16',
            ),
            [f.MaxBytes.CODE_TOO_LONG],

            # End result is only 12 bytes because UTF-16 uses 2 bytes
            # per character.
            #
            # Technically, it's only 10 bytes if you don't count the
            # BOM.
            expected_value =
                b'\xff\xfe\x93\x03\xb5\x03\xb9\x03\xac\x03\xc3\x03',
        )

    def test_fail_string_alt_encoding_with_prefix(self):
        """
        The filter is configured to use an encoding other than UTF-8
        and apply a prefix to truncated values.
        """
        self.assertFilterErrors(
            self._filter(
                'Γειάσου Κόσμε',

                max_bytes   = 18,
                prefix      = 'σφάλμα:',
                encoding    = 'utf-16',
            ),
            [f.MaxBytes.CODE_TOO_LONG],

            # Note that the BOM is only applied once.
            expected_value =
                # BOM:
                b'\xff\xfe'
                # Prefix:
                b'\xc3\x03\xc6\x03\xac\x03\xbb'
                b'\x03\xbc\x03\xb1\x03:\x00'
                # Truncated string:
                b'\x93\x03',
        )

    def test_pass_bytes_short(self):
        """
        The incoming value is a byte string that is short enough.
        """
        self.assertFilterPasses(
            self._filter(
                b'\xe4\xbd\xa0\xe5\xa5\xbd\xef\xbc\x8c'
                b'\xe4\xb8\x96\xe7\x95\x8c\xef\xbc\x81',

                max_bytes = 18,
            ),
        )

    def test_fail_bytes_long(self):
        """
        The incoming value is a byte string that is too long.
        """
        self.assertFilterErrors(
            self._filter(
                b'\xe4\xbd\xa0\xe5\xa5\xbd\xef\xbc\x8c'
                b'\xe4\xb8\x96\xe7\x95\x8c\xef\xbc\x81',

                max_bytes   = 17,
            ),
            [f.MaxBytes.CODE_TOO_LONG],

            # Note that the resulting value is truncated to 15 bytes
            # instead of 17, so as not to orphan a multibyte
            # sequence.
            expected_value =
                b'\xe4\xbd\xa0\xe5\xa5\xbd\xef'
                b'\xbc\x8c\xe4\xb8\x96\xe7\x95\x8c',
        )

    def test_fail_wrong_type(self):
        """
        The incoming value is not a string.
        """
        self.assertFilterErrors(
            self._filter(['foo', 'bar'], max_bytes=32),
            [f.Type.CODE_WRONG_TYPE],
        )


class RegexTestCase(BaseFilterTestCase):
    filter_type = f.Regex

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use ``Required | Regex`` if you want to reject null values.
        """
        self.assertFilterPasses(self._filter(None, pattern=r'.'))

    def test_pass_happy_path(self):
        """
        The incoming value matches the regex pattern.
        """
        # noinspection SpellCheckingInspection
        self.assertFilterPasses(
            # Note:  Regexes are case-sensitive by default.
            self._filter(
                'test march of the TEST penguins',
                pattern = r'\btest\b',
            ),

            ['test'],
        )

    def test_fail_no_match(self):
        """
        The incoming value does not match the regex pattern.
        """
        # noinspection SpellCheckingInspection
        self.assertFilterErrors(
            self._filter(
                'contested march of the tester penguins',
                pattern=r'\btest\b',
            ),

            [f.Regex.CODE_INVALID],
        )

    def test_pass_unicode_character_class(self):
        """
        By default, character classes like ``\w`` will take unicode into
        account.
        """
        # Roughly, "Hi there!" in Japanese.
        word = '\u304a\u306f\u3088\u3046'

        self.assertFilterPasses(
            self._filter(word, pattern=r'\w+'),
            [word],
        )

    def test_pass_precompiled_regex(self):
        """
        You can alternatively provide a precompiled regex to the Filter
        instead of a string pattern.
        """
        # Compile our own pattern so that we can specify the
        # ``IGNORECASE`` flag.
        # Note that you are responsible for adding the ``UNICODE`` flag
        # to your compiled regex!
        # noinspection SpellCheckingInspection
        pattern = re.compile(r'\btest\b', re.IGNORECASE | re.UNICODE)

        self.assertFilterPasses(
            self._filter('test march of the TEST penguins', pattern=pattern),
            ['test', 'TEST'],
        )

    def test_pass_regex_library_support(self):
        """
        The Regex Filter also supports precompiled patterns using the
        ``regex`` library.
        """
        # Roughly, "Hi there!" in Burmese.
        word = '\u101f\u102d\u102f\u1004\u103a\u1038'

        # Note that :py:func:`regex.compile` automatically adds the
        # ``UNICODE`` flag for you when the pattern is a unicode.
        pattern = regex.compile(r'\w+')

        self.assertFilterPasses(
            self._filter(word, pattern=pattern),
            [word],
        )

    def test_fail_bytes(self):
        """
        To ensure consistent behavior in Python 2 and Python 3, byte
        strings are not allowed.
        """
        self.assertFilterErrors(
            self._filter(b"Aw, come on; it'll be fun!", pattern=r'.'),
            [f.Type.CODE_WRONG_TYPE],
        )

    def test_fail_wrong_type(self):
        """
        The incoming value is not a string.
        """
        self.assertFilterErrors(
            self._filter(['totally', 'valid', 'right?'], pattern=r'.'),
            [f.Type.CODE_WRONG_TYPE],
        )


class SplitTestCase(BaseFilterTestCase):
    filter_type = f.Split

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use ``Required | Split`` if you want to reject null values.
        """
        self.assertFilterPasses(
            self._filter(None, pattern='test'),
        )

    def test_pass_char_split(self):
        """
        Simplest use case is to split a string by a single character.
        """
        self.assertFilterPasses(
            self._filter('foo:bar:baz', pattern=':'),
            ['foo', 'bar', 'baz'],
        )

    def test_pass_pattern_split(self):
        """
        You can also use a regex to split the string.
        """
        self.assertFilterPasses(
            self._filter('foo-12-bar-34-baz', pattern='[-\d]+'),
            ['foo', 'bar', 'baz'],
        )

    def test_pass_pattern_split_with_groups(self):
        """
        If you include capturing parentheses in the pattern, the
        matched groups are included in the resulting list.
        """
        self.assertFilterPasses(
            # Note grouping parentheses in the regex.
            self._filter('foo-12-bar-34-baz', pattern='([-\d]+)'),
            ['foo', '-12-', 'bar', '-34-', 'baz'],
        )

    def test_pass_no_split(self):
        """
        A string that does not match the regex at all is also
        considered valid.

        Use ``Split | MinLength`` if you want to require a minimum
        number of parts.
        """
        self.assertFilterPasses(
            self._filter('foo:bar:baz', pattern='[-\d]+'),
            ['foo:bar:baz'],
        )

    def test_pass_keys(self):
        """
        If desired, you can map a collection of keys onto the resulting
        list, which creates an OrderedDict.

        This is particularly cool, as it lets you chain a Split with a
        FilterMapper.
        """
        filtered = self._filter(
            'foo:bar:baz',
                pattern = ':',
                keys    = ('a', 'b', 'c',),
        )

        self.assertFilterPasses(filtered, self.skip_value_check)

        cleaned = filtered.cleaned_data
        self.assertIsInstance(cleaned, OrderedDict)

        self.assertDictEqual(cleaned, {
            'a':    'foo',
            'b':    'bar',
            'c':    'baz',
        })

        # Because the result is an OrderedDict, the order is preserved
        # as well.
        self.assertListEqual(
            list(itervalues(cleaned)),
            ['foo', 'bar', 'baz'],
        )

    def test_pass_precompiled_regex(self):
        """
        You can alternatively provide a precompiled regex to the Filter
        instead of a string pattern.
        """
        # Compile our own pattern so that we can specify the
        # ``IGNORECASE`` flag.
        # Note that you are responsible for adding the ``UNICODE`` flag
        # to your compiled regex!
        # noinspection SpellCheckingInspection
        pattern = re.compile(r'\btest\b', re.IGNORECASE | re.UNICODE)

        self.assertFilterPasses(
            self._filter('test march of the TEST penguins', pattern=pattern),
            ['', ' march of the ', ' penguins'],
        )

    def test_pass_regex_library_support(self):
        """
        The Regex Filter also supports precompiled patterns using the
        ``regex`` library.
        """
        # Roughly, "Hi there!" in Burmese.
        word = '\u101f\u102d\u102f\u1004\u103a\u1038!'

        # Note that :py:func:`regex.compile` automatically adds the
        # ``UNICODE`` flag for you when the pattern is a unicode.
        pattern = regex.compile(r'\w+')

        self.assertFilterPasses(
            self._filter(word, pattern=pattern),
            ['', '!'],
        )

    def test_fail_too_long(self):
        """
        The incoming value has too many parts to assign a key to each
        one, so it fails validation.
        """
        self.assertFilterErrors(
            self._filter('foo:bar:baz', pattern=':', keys=('a', 'b',)),
            [f.MaxLength.CODE_TOO_LONG],
        )

    def test_pass_too_short(self):
        """
        The incoming value does not have enough parts to use all the
        keys, so extra ``None`` values are inserted.

        If you want to ensure that the incoming value has exactly the
        right number of values, add a MinLength filter to the chain
        (you do not have to provide a MaxLength; the Split filter
        does that automatically).
        """
        self.assertFilterPasses(
            self._filter(
                'foo:bar:baz',
                    pattern = ':',
                    keys    = ('a', 'b', 'c', 'd',),
            ),

            {
                'a':    'foo',
                'b':    'bar',
                'c':    'baz',
                'd':    None,
            },
        )

    def test_fail_bytes(self):
        """
        To ensure consistent behavior in Python 2 and Python 3, byte
        strings are not allowed.
        """
        self.assertFilterErrors(
            self._filter(b'foo bar baz', pattern=''),
            [f.Type.CODE_WRONG_TYPE],
        )

    def test_fail_wrong_type(self):
        """
        The incoming value is not a string.
        """
        self.assertFilterErrors(
            self._filter(['foo', 'bar', 'baz'], pattern=''),
            [f.Type.CODE_WRONG_TYPE],
        )


class StripTestCase(BaseFilterTestCase):
    filter_type = f.Strip

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use `Required | Strip` if you want to reject null values.
        """
        self.assertFilterPasses(None)

    def test_pass_happy_path(self):
        """
        The Filter strips away all leading/trailing whitespace and
        unprintables from the incoming value.
        """
        self.assertFilterPasses(
            '  \r  \t \x00 Hello, world! \x00 \t  \n  ',
            'Hello, world!',
        )

    def test_pass_leading_only(self):
        """
        You can configure the filter to strip leading characters only.
        """
        self.assertFilterPasses(
            self._filter(
                '  \r  \t \x00 Hello, world! \x00 \t  \n  ',
                    trailing = None,
            ),

            'Hello, world! \x00 \t  \n  ',
        )

    def test_pass_trailing_only(self):
        """
        You can configure the filter to strip trailing characters only.
        """
        self.assertFilterPasses(
            self._filter(
                '  \r  \t \x00 Hello, world! \x00 \t  \n  ',
                    leading = None,
            ),
            '  \r  \t \x00 Hello, world!',
        )

    def test_pass_unicode(self):
        """
        Strip also catches non-ASCII characters that are classified as
        whitespace according to Unicode.
        """
        # U+2003 is an em space.
        self.assertFilterPasses(
            '\u2003Hello, world!\u2003',
            'Hello, world!',
        )

    def test_pass_custom_regexes(self):
        """
        You can also use regexes to specify which characters get
        removed.
        """
        self.assertFilterPasses(
            self._filter(
                "54321 Hello, world! "
                        "i think you ought to know i'm feeling very depressed ",
                    leading     = r'\d',
                    trailing    = r"['a-z ]+"
            ),

            '4321 Hello, world!',
        )

    def test_fail_bytes(self):
        """
        To ensure consistent behavior in Python 2 and Python 3, byte
        strings are not allowed.
        """
        self.assertFilterErrors(
            b'    but... but... look at all of my whitespace!    ',
            [f.Type.CODE_WRONG_TYPE],
        )

    def test_fail_wrong_type(self):
        """
        The incoming value is not a string.
        """
        self.assertFilterErrors(
            ['  lots  ', '  of  ', '  whitespace  ', '  here  '],
            [f.Type.CODE_WRONG_TYPE],
        )


class UuidTestCase(BaseFilterTestCase):
    filter_type = f.Uuid

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use ``Required | Uuid`` if you want to reject null values.
        """
        self.assertFilterPasses(None)

    def test_pass_uuid_value(self):
        """
        The incoming value can be interpreted as a UUID.
        """
        filtered = self._filter('3466c56a-2ebc-449d-97d2-9b119721ff0f')

        self.assertFilterPasses(filtered, self.skip_value_check)

        uuid = filtered.cleaned_data
        self.assertIsInstance(uuid, UUID)

        self.assertEqual(uuid.hex, '3466c56a2ebc449d97d29b119721ff0f')
        self.assertEqual(uuid.version, 4)

    def test_pass_hex(self):
        """
        You can omit the dashes when specifying a UUID value.
        """
        filtered = self._filter('3466c56a2ebc449d97d29b119721ff0f')

        self.assertFilterPasses(filtered, self.skip_value_check)

        uuid = filtered.cleaned_data
        self.assertIsInstance(uuid, UUID)

        self.assertEqual(uuid.hex, '3466c56a2ebc449d97d29b119721ff0f')
        self.assertEqual(uuid.version, 4)

    # noinspection SpellCheckingInspection
    def test_pass_curly_hex(self):
        """
        You can include curly braces around hex values.

        Use ``Regex(r'^[\da-f]+$') | Uuid`` if you only want to allow
        plain hex.
        """
        filtered = self._filter('{54d6ebf8a3f55ed59becdedfb3b0773f}')

        self.assertFilterPasses(filtered, self.skip_value_check)

        uuid = filtered.cleaned_data
        self.assertIsInstance(uuid, UUID)

        self.assertEqual(uuid.hex, '54d6ebf8a3f55ed59becdedfb3b0773f')
        self.assertEqual(uuid.version, 5)

    def test_pass_urn(self):
        """
        You can also specify a URN.  The term (and format) is somewhat
        antiquated, but still valid.

        If you want to prohibit URNs, chain this Filter with
        ``Regex(r'^[\da-f]+$')``.

        References:
          - https://en.wikipedia.org/wiki/Uniform_resource_name
        """
        filtered = self._filter(
            'urn:uuid:2830f705-5969-11e5-9628-e0f8470933c8',
        )

        self.assertFilterPasses(filtered, self.skip_value_check)

        uuid = filtered.cleaned_data
        self.assertIsInstance(uuid, UUID)

        self.assertEqual(uuid.hex, '2830f705596911e59628e0f8470933c8')
        self.assertEqual(uuid.version, 1)

    def test_fail_wrong_version(self):
        """
        The incoming value is a valid UUID, but its version doesn't
        match the expected one.
        """
        self.assertFilterErrors(
            # Incoming value is a v1 UUID, but we're expecting a v4.
            self._filter('2830f705596911e59628e0f8470933c8', version=4),
            [f.Uuid.CODE_WRONG_VERSION],
        )

    def test_fail_int(self):
        """
        The incoming value must be a HEX representation of a UUID.
        Decimal values are not valid.
        """
        self.assertFilterErrors(
            '306707680894066278898485957190279549189',
            [f.Uuid.CODE_INVALID],
        )

    def test_fail_wrong_type(self):
        """
        Attempting to Filter anything other than a string value fails
        rather spectacularly.
        """
        self.assertFilterErrors(
            [
                'e6bdc02c9d004991986d3c7c0730d105',
                '2830f705596911e59628e0f8470933c8',
            ],

            [f.Type.CODE_WRONG_TYPE],
        )

    def test_pass_uuid_object(self):
        """
        The incoming value is already a UUID object.
        """
        self.assertFilterPasses(UUID('e6bdc02c9d004991986d3c7c0730d105'))

    def test_fail_uuid_object_wrong_version(self):
        """
        The incoming value is already a UUID object, but its version
        doesn't match the expected one.
        """
        # noinspection SpellCheckingInspection
        self.assertFilterErrors(
            # Incoming value is a v5 UUID, but we're expecting a v4.
            self._filter(UUID('54d6ebf8a3f55ed59becdedfb3b0773f'), version=4),
            [f.Uuid.CODE_WRONG_VERSION],
        )


class UnicodeTestCase(BaseFilterTestCase):
    filter_type = f.Unicode

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use ``Required | Unicode`` if you want to reject null values.
        """
        self.assertFilterPasses(None)

    def test_pass_unicode(self):
        """
        The incoming value is a unicode.
        """
        self.assertFilterPasses('┻━┻︵ \(°□°)/ ︵ ┻━┻ ') # RAWR!

    def test_pass_bytes_utf8(self):
        """
        The incoming value is a byte string that is encoded as UTF-8.
        """
        self.assertFilterPasses(
            # You get used to it.
            # I don't even see the code.
            # All I see is, "blond"... "brunette"... "redhead"...
            # Hey, you uh... want a drink?
            b'\xe2\x99\xaa '
            b'\xe2\x94\x8f(\xc2\xb0.\xc2\xb0)\xe2\x94\x9b '
            b'\xe2\x94\x97(\xc2\xb0.\xc2\xb0)\xe2\x94\x93 '
            b'\xe2\x99\xaa',

            '♪ ┏(°.°)┛ ┗(°.°)┓ ♪',
        )

    def test_fail_bytes_non_utf8(self):
        """
        The incoming value is a byte string that is encoded using a
        codec other than UTF-8.

        Note that there is no such thing as a unicode object with
        the "wrong encoding".

        References:
          - https://docs.python.org/2/howto/unicode.html
        """
        # How about something a bit more realistic for this test?
        # Like the Swedish word for 'Apple'.
        # noinspection SpellCheckingInspection
        incoming = b'\xc4pple'

        self.assertFilterErrors(
            incoming,
            [f.Unicode.CODE_DECODE_ERROR],
        )

        # In order for this to work, we have to tell the Filter what
        # encoding to use:
        self.assertFilterPasses(
            self._filter(incoming, encoding='iso-8859-1'),
            'Äpple',
        )

    def test_pass_string_like_object(self):
        """
        The incoming value is an object that can be cast as a string.
        """
        value = '／人 ⌒ ‿‿ ⌒ 人＼' # Squee!

        self.assertFilterPasses(
            Unicody(value),
            value,
        )

    def test_pass_bytes_like_object(self):
        """
        The incoming value is an object that can be cast as a byte
        string.
        """
        self.assertFilterPasses(
            Bytesy(b'(\xe2\x99\xa5\xe2\x80\xbf\xe2\x99\xa5)'),

            # I can almost hear the sappy music now.
            '(♥‿♥)',
        )

    def test_pass_boolean(self):
        """
        The incoming value is a boolean (treated as an int).
        """
        self.assertFilterPasses(True, '1')

    def test_pass_decimal_with_scientific_notation(self):
        """
        The incoming value is a Decimal that was parsed from scientific
        notation.
        """
        # Note that `text_type(Decimal('2.8E6')` yields '2.8E+6', which
        # is not what we want!
        self.assertFilterPasses(
            Decimal('2.8E6'),
            '2800000',
        )

    def test_pass_xml_element(self):
        """
        The incoming value is an ElementTree XML Element.
        """
        self.assertFilterPasses(
            Element('foobar'),
            '<foobar />',
        )

    def test_unicode_normalization(self):
        """
        The Filter always returns the NFC form of the unicode string.

        References:
          - https://en.wikipedia.org/wiki/Unicode_equivalence
          - http://stackoverflow.com/q/16467479
        """
        #   U+0065 LATIN SMALL LETTER E
        # + U+0301 COMBINING ACUTE ACCENT
        # (2 code points)
        decomposed  = 'Ame\u0301lie'

        # U+00E9 LATIN SMALL LETTER E WITH ACUTE
        # (1 code point)
        composed    = 'Am\xe9lie'

        self.assertFilterPasses(decomposed, composed)

    def test_unicode_normalization_disabled(self):
        """
        You can force the Filter not to perform normalization.
        """
        decomposed  = 'Ame\u0301lie'

        self.assertFilterPasses(
            self._filter(decomposed, normalize=False),
            decomposed,
        )

    def test_remove_non_printables(self):
        """
        By default, this Filter also removes non-printable characters
        (both ASCII and Unicode varieties).
        """
        self.assertFilterPasses(
            # \x00-\x1f are ASCII control characters.
            # \xef\xbf\xbf is the Unicode control character \uffff,
            # encoded as UTF-8.
            b'\x10Hell\x00o,\x1f wor\xef\xbf\xbfld!',

            'Hello, world!',
        )

    def test_remove_non_printables_disabled(self):
        """
        You can force the Filter not to remove non-printable characters.
        """
        self.assertFilterPasses(
            self._filter(
                b'\x10Hell\x00o,\x1f wor\xef\xbf\xbfld!',
                normalize = False,
            ),

            '\u0010Hell\u0000o,\u001f wor\uffffld!',
        )

    def test_newline_normalization(self):
        """
        By default, any newlines in the string are automatically
        converted to unix-style ('\n').
        """
        self.assertFilterPasses(
            b'unix\n - windows\r\n - weird\r\r\n',
            'unix\n - windows\n - weird\n\n',
        )

    def test_newline_normalization_disabled(self):
        """
        You can force the Filter not to normalize line endings.
        """
        self.assertFilterPasses(
            self._filter(
                b'unix\n - windows\r\n - weird\r\r\n',
                normalize = False,
            ),

            'unix\n - windows\r\n - weird\r\r\n',
        )



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from collections import Sequence
from datetime import date, datetime
from typing import Sized

from dateutil.tz import tzoffset
from pytz import utc
from six import PY2, binary_type, python_2_unicode_compatible, text_type

import filters as f
from filters.test import BaseFilterTestCase


class Lengthy(Sized):
    """
    A class that defines ``__len__``, used to test Filters that check
    for object length.
    """
    def __init__(self, length):
        super(Lengthy, self).__init__()
        self.length = length

    def __len__(self):
        return self.length


# noinspection SpellCheckingInspection
class Bytesy(object):
    """
    A class that defines ``__bytes__``, used to test Filters that
    convert values into byte strings.
    """
    def __init__(self, value):
        super(Bytesy, self).__init__()
        self.value = value

    def __bytes__(self):
        return binary_type(self.value)

    if PY2:
        __str__ = __bytes__


# noinspection SpellCheckingInspection
@python_2_unicode_compatible
class Unicody(object):
    """
    A class that defines ``__str__``, used to test Filters that convert
    values into unicodes.
    """
    def __init__(self, value):
        super(Unicody, self).__init__()
        self.value = value

    def __str__(self):
        return text_type(self.value)


class ArrayTestCase(BaseFilterTestCase):
    filter_type = f.Array

    def test_pass_none(self):
        """
        ``None`` always passes this filter.

        Use ``Required | Array`` if you want to reject null values.
        """
        self.assertFilterPasses(None)

    def test_pass_sequence(self):
        """
        The incoming value is a sequence.
        """
        self.assertFilterPasses(tuple())
        self.assertFilterPasses(list())

    def test_pass_custom_sequence_type(self):
        """
        The incoming value has a type that extends Sequence.
        """
        class CustomSequence(Sequence):
            """
            Technically, it's a Sequence. Technically.
            """
            def __len__(self): return 0
            def __getitem__(self, index): return None

        self.assertFilterPasses(CustomSequence())

    def test_fail_string(self):
        """
        The incoming value is a string.
        """
        self.assertFilterErrors(binary_type(), [f.Array.CODE_WRONG_TYPE])
        self.assertFilterErrors(text_type(), [f.Array.CODE_WRONG_TYPE])

    def test_fail_mapping(self):
        """
        The incoming value is a mapping.
        """
        self.assertFilterErrors(dict(), [f.Array.CODE_WRONG_TYPE])

    def test_fail_set(self):
        """
        The incoming value is a set.
        """
        self.assertFilterErrors(set(), [f.Array.CODE_WRONG_TYPE])

    def test_fail_custom_sequence_type(self):
        """
        The incoming value looks like a Sequence, but it's not official.
        """
        class CustomSequence(object):
            """
            Walks, talks and quacks like a Sequence, but isn't.
            """
            def __len__(self): return 0
            def __getitem__(self, index): return None

        self.assertFilterErrors(CustomSequence(), [f.Array.CODE_WRONG_TYPE])

        # If you can't (or don't want) to modify the base class for
        # your custom sequence, you can register it.
        # Note: Code included here for documentation purposes, but it's
        # commented out to avoid side effects; registering a subclass
        # this way is basically irreversible.
        # Sequence.register(CustomSequence)
        # self.assertFilterPasses(CustomSequence())


class ByteArrayTestCase(BaseFilterTestCase):
    filter_type = f.ByteArray

    def test_pass_none(self):
        """
        ``None`` always passes this filter.

        Use ``Required | ByteArray`` if you want to reject null values.
        """
        self.assertFilterPasses(None)

    def test_pass_bytes(self):
        """
        The incoming value is a byte string.
        """
        self.assertFilterPasses(
            b'|\xa8\xc1.8\xbd4\xd5s\x1e\xa6%+\xea!6',

            # Note that "numeric" characters like "8" and "6" are NOT
            # interpreted literally.
            # This matches the behavior of Python's built-in bytearray
            # class.
            bytearray([
                124, 168, 193, 46, 56, 189, 52, 213,
                115, 30, 166, 37, 43, 234, 33, 54,
            ]),
        )

    def test_pass_string(self):
        """
        The incoming value is a string.

        This is generally not a recommended use for ByteArray, but
        sometimes it's unavoidable.
        """
        self.assertFilterPasses(
            u'\xccK\xdf\xb1\x8bM\xc7\x01\xf0B\xac":\xeb>\x85',

            bytearray([
                195, 140, 75, 195, 159, 194, 177, 194, 139, 77, 195, 135,
                1, 195, 176, 66, 194, 172, 34, 58, 195, 171, 62, 194, 133,
            ]),
        )

    def test_pass_string_alternate_encoding(self):
        """
        If you want to filter unicodes, you can specify the encoding to
        use.
        """
        self.assertFilterPasses(
            self._filter(
                u'\xccK\xdf\xb1\x8bM\xc7\x01\xf0B\xac":\xeb>\x85',
                encoding='latin-1',
            ),

            bytearray([
                204, 75, 223, 177, 139, 77, 199, 1,
                240, 66, 172, 34, 58, 235, 62, 133,
            ]),
        )

    def test_pass_bytearray(self):
        """
        The incoming value is already a bytearray.
        """
        self.assertFilterPasses(
            bytearray([
                84, 234, 48, 177, 119, 69, 36, 147,
                214, 13, 54, 12, 56, 168, 107, 2,
            ])
        )

    def test_pass_iterable(self):
        """
        The incoming value is an iterable containing integers between
        0 and 255, inclusive.
        """
        self.assertFilterPasses(
            [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89, 144, 233],
            bytearray([0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89, 144, 233])
        )

    def test_fail_iterable_wrong_types(self):
        """
        The incoming value is an iterable, but the values are not
        integers.
        """
        self.assertFilterErrors(
            # The first 2 values are valid.  None of the others
            # are.
            # It's arguable whether booleans should be valid, but
            # they are technically ints, and Python's bytearray
            # allows them, so the Filter does, too.
            [1, True, '1', b'1', 1.1, bytearray([1])],

            {
                #
                # String values inside of an iterable are not
                # considered valid.
                #
                # It's true that we do have a precedent for how to
                # treat string values (convert each character to
                # its ordinal value), but that only works for
                # strings that can fit into a single byte.
                #
                # How would we convert `['11', 'foo']` into a
                # bytearray?
                #
                # To keep things as consistent as possible, the
                # Filter will treat strings inside of iterables
                # the same way it treats anything else that isn't
                # an int.
                #
                '2': [f.Type.CODE_WRONG_TYPE],
                '3': [f.Type.CODE_WRONG_TYPE],

                # Floats are not allowed in bytearrays.  How would
                # that even work?
                '4': [f.Type.CODE_WRONG_TYPE],

                # Anything else that isn't an int is invalid, even
                # if it contains ints.
                # After all, you can't squeeze multiple bytes into
                # a single byte!
                '5': [f.Type.CODE_WRONG_TYPE],
            },
        )

    def test_fail_iterable_out_of_bounds(self):
        """
        The incoming value is an iterable with integers, but it
        contains values outside the acceptable range.

        Each value inside a bytearray must fit within 1 byte, so its
        value must satisfy ``0 <= x < 2^8``.
        """
        self.assertFilterErrors(
            [-1, 0, 1, 255, 256, 9001],

            {
                '0': [f.Min.CODE_TOO_SMALL],
                '4': [f.Max.CODE_TOO_BIG],
                '5': [f.Max.CODE_TOO_BIG],
            },
        )

    def test_fail_unencodable_unicode(self):
        """
        The incoming value is a unicode that cannot be encoded using
        the specified encoding.
        """
        value = '\u043b\u0435\u0431\u044b\u0440'

        # The default encoding (utf-8) can handle this just fine.
        self.assertFilterPasses(
            value,
            bytearray([208, 187, 208, 181, 208, 177, 209, 139, 209, 128]),
        )

        # However, if we switch to a single-byte encoding, we run into
        # serious problems.
        self.assertFilterErrors(
            self._filter(value, encoding='latin-1'),
            [f.ByteArray.CODE_BAD_ENCODING],
        )


class ChoiceTestCase(BaseFilterTestCase):
    filter_type = f.Choice

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use ``Required | Choice`` if you want to reject null values.
        """
        self.assertFilterPasses(
            # Even if you specify no valid choices, `None` still
            # passes.
            self._filter(None, choices=()),
        )

    def test_pass_valid_value(self):
        """
        The incoming value matches one of the choices.
        """
        self.assertFilterPasses(
            self._filter('Curly', choices=('Moe', 'Larry', 'Curly')),
        )

    def test_fail_invalid_value(self):
        """
        The incoming value does not match any of the choices.
        """
        self.assertFilterErrors(
            self._filter('Shemp', choices=('Moe', 'Larry', 'Curly')),
            [f.Choice.CODE_INVALID],
        )


class DateTestCase(BaseFilterTestCase):
    filter_type = f.Date

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use `Required | Date` if you want to reject null values.
        """
        self.assertFilterPasses(None)

    def test_pass_naive_timestamp(self):
        """
        The incoming value is a naive timestamp (no timezone info).
        """
        self.assertFilterPasses(
            '2015-05-11 14:56:58',
            date(2015, 5, 11),
        )

    def test_pass_aware_timestamp(self):
        """
        The incoming value includes timezone info.
        """
        self.assertFilterPasses(
            # Note that the value we are parsing is 5 hours behind UTC.
            '2015-05-11T19:56:58-05:00',

            # The resulting date appears to occur 1 day later because
            # that's the date according to UTC.
            date(2015, 5, 12)
        )

    def test_pass_naive_timestamp_default_timezone(self):
        """
        The incoming value is a naive timestamp, but the Filter is
        configured not to treat naive timestamps as UTC.
        """
        self.assertFilterPasses(
            self._filter(
                '2015-05-12 03:20:03',

                # The Filter is configured to parse naive timestamps as
                # if they are UTC+8.
                timezone = tzoffset('UTC+8', 8*3600)
            ),

            # The resulting date appears to occur 1 day earlier because
            # the Filter subtracted 8 hours to convert the value to
            # UTC.
            date(2015, 5, 11),
        )

    def test_pass_aware_timestamp_default_timezone(self):
        """
        The Filter's default timezone has no effect if the incoming
        value already contains timezone info.
        """
        self.assertFilterPasses(
            # The incoming timestamp is from UTC+4, but the Filter is
            # configured to use UTC-11 by default.
            self._filter(
                '2015-05-11T03:14:38+04:00',
                timezone = tzoffset('UTC-11', -11*3600)
            ),

            # Because the incoming timestamp has timezone info, the
            # Filter uses that instead of the default value.
            # Note that the this test will fail if the Filter uses the
            # UTC-11 timezone (the result will be 1 day ahead).
            date(2015, 5, 10),
        )

    def test_pass_alternate_timezone_syntax(self):
        """
        When setting the default timezone for the Filter, you can use
        an int/float offset (number of hours from UTC) instead of a
        tzoffset object.
        """
        self.assertFilterPasses(
            # Note that we use an int value instead of constructing a
            # tzoffset for `timezone`.
            self._filter('2015-05-11 21:14:38', timezone=-8),
            date(2015, 5, 12),
        )

    def test_pass_datetime_utc(self):
        """
        The incoming value is a datetime object that is already set to
        UTC.
        """
        self.assertFilterPasses(
            datetime(2015, 6, 27, 10, 5, 48, tzinfo=utc),
            date(2015, 6, 27),
        )

    def test_pass_datetime_non_utc(self):
        """
        The incoming value is a datetime object with a non-UTC
        timezone.
        """
        self.assertFilterPasses(
            datetime(
                2015, 6, 27, 22, 6, 32,
                tzinfo=tzoffset('UTC-5', -5*3600),
            ),

            # As you probably already guessed, the datetime gets
            # converted to UTC before it is converted to a date.
            date(2015, 6, 28),
        )

    def test_pass_datetime_naive(self):
        """
        The incoming value is a datetime object without timezone info.
        """
        self.assertFilterPasses(
            # The Filter will assume that this datetime is UTC-3 by
            # default.
            self._filter(datetime(2015, 6, 27, 23, 7, 18), timezone=-3),

            # The datetime is converted from UTC-3 to UTC before it is
            # converted to a date.
            date(2015, 6, 28),
        )

    def test_pass_date(self):
        """
        The incoming value is a date object.
        """
        self.assertFilterPasses(date(2015, 6, 27))

    def test_fail_invalid_value(self):
        """
        The incoming value cannot be interpreted as a date.

        Insert socially-awkward nerd joke here.
        """
        self.assertFilterErrors(
            'this is not a date', # it's a space station
            [f.Date.CODE_INVALID],
        )


class DatetimeTestCase(BaseFilterTestCase):
    filter_type = f.Datetime

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use `Required | Datetime` if you want to reject null values.
        """
        self.assertFilterPasses(None)

    def test_pass_naive_timestamp(self):
        """
        The incoming value is a naive timestamp (does not include
        timezone info).
        """
        self.assertFilterPasses(
            '2015-05-11 14:56:58',
            datetime(2015, 5, 11, 14, 56, 58, tzinfo=utc),
        )

    def test_pass_aware_timestamp(self):
        """
        The incoming value is a timestamp that includes timezone info.
        """
        self.assertFilterPasses(
            # Note that the value we are parsing is 5 hours behind UTC.
            '2015-05-11T14:56:58-0500',

            datetime(2015, 5, 11, 19, 56, 58, tzinfo=utc)
        )

    def test_pass_naive_timestamp_default_timezone(self):
        """
        The incoming value is a naive timestamp, but the Filter is
        configured not to treat naive timestamps as UTC.
        """
        self.assertFilterPasses(
            # The incoming value is a naive timestamp, and the Filter
            # is configured to use UTC+8 by default.
            self._filter(
                '2015-05-12 09:20:03',
                timezone = tzoffset('UTC+8', 8*3600),
            ),

            # The resulting datetime is still converted to UTC.
            datetime(2015, 5, 12, 1, 20, 3, tzinfo=utc),
        )

    def test_pass_aware_timestamp_default_timezone(self):
        """
        The Filter's default timezone has no effect if the incoming
        value already contains timezone info.
        """
        self.assertFilterPasses(
            # The incoming value is UTC+4, but the Filter is configured
            # to use UTC-1 by default.
            self._filter(
                '2015-05-11T21:14:38+04:00',
                timezone = tzoffset('UTC-1', -1*3600)
            ),

            # The incoming values timezone info is used instead of the
            # default.
            # Note that the resulting datetime is still converted to
            # UTC.
            datetime(2015, 5, 11, 17, 14, 38, tzinfo=utc),
        )

    def test_pass_alternate_timezone_syntax(self):
        """
        When setting the default timezone for the Filter, you can use
        an int/float offset (number of hours from UTC) instead of a
        tzoffset object.
        """
        self.assertFilterPasses(
            # Note that we use an int value instead of constructing a
            # tzoffset for ``timezone``.
            self._filter('2015-05-11 21:14:38', timezone=3),

            datetime(2015, 5, 11, 18, 14, 38, tzinfo=utc),
        )

    def test_pass_datetime_utc(self):
        """
        The incoming value is a datetime object that is already set to
        UTC.
        """
        self.assertFilterPasses(datetime(2015, 6, 27, 10, 5, 48, tzinfo=utc))

    def test_pass_datetime_non_utc(self):
        """
        The incoming value is a datetime object that is already set to
        a non-UTC timezone.
        """
        self.assertFilterPasses(
            datetime(2015, 6, 27, 10, 6, 32, tzinfo=tzoffset('UTC-5',-5*3600)),
            datetime(2015, 6, 27, 15, 6, 32, tzinfo=utc),
        )

    def test_datetime_naive(self):
        """
        The incoming value is a datetime object that does not have
        timezone info.
        """
        self.assertFilterPasses(
            # The Filter is configured to assume UTC-3 if the incoming
            # value has no timezone info.
            self._filter(datetime(2015, 6, 27, 18, 7, 18), timezone=-3),

            datetime(2015, 6, 27, 21, 7, 18, tzinfo=utc),
        )

    def test_pass_date(self):
        """
        The incoming value is a date object.
        """
        self.assertFilterPasses(
            # The Filter is configured to assume UTC+12 if the incoming
            # value has no timezone info.
            self._filter(date(2015, 6, 27), timezone=12),

            datetime(2015, 6, 26, 12, 0, 0, tzinfo=utc),
        )

    def test_return_naive_datetime(self):
        """
        You can configure the filter to return a naive datetime object
        (e.g., for storing in a database).

        Note that the datetime is still converted to UTC before its
        tzinfo is removed.
        """
        self.assertFilterPasses(
            self._filter(
                datetime(
                    2015, 7, 1, 9, 22, 10,
                    tzinfo=tzoffset('UTC-5', -5*3600),
                ),

                # Note that we pass `naive=True` to the Filter's
                # initializer.
                naive = True,
            ),

            # The resulting datetime is converted to UTC before its
            # timezone info is stripped.
            datetime(2015, 7, 1, 14, 22, 10, tzinfo=None),
        )

    def test_fail_invalid_value(self):
        """
        The incoming value cannot be parsed as a datetime.
        """
        self.assertFilterErrors(
            'this is not a datetime', # it's a pipe
            [f.Datetime.CODE_INVALID],
        )


class EmptyTestCase(BaseFilterTestCase):
    filter_type = f.Empty

    def test_pass_none(self):
        """
        ``None`` shall pass.

        What?

        ``None`` shall pass!
        """
        self.assertFilterPasses(None)

    def test_pass_empty_string(self):
        """
        The incoming value is an empty string.
        """
        self.assertFilterPasses('')

    def test_pass_empty_collection(self):
        """
        The incoming value is a collection with length < 1.
        """
        self.assertFilterPasses([])
        self.assertFilterPasses({})
        self.assertFilterPasses(Lengthy(0))
        # etc.

    def test_fail_non_empty_string(self):
        """
        The incoming value is a non-empty string.
        """
        self.assertFilterErrors(
            'Goodbye world!',
            [f.Empty.CODE_NOT_EMPTY],
        )

    def test_fail_non_empty_collection(self):
        """
        The incoming value is a collection with length > 0.
        """
        # The values inside the collection may be empty, but the
        # collection itself is not.
        self.assertFilterErrors(['', '', ''],   [f.Empty.CODE_NOT_EMPTY])
        self.assertFilterErrors({'': ''},       [f.Empty.CODE_NOT_EMPTY])
        self.assertFilterErrors(Lengthy(1),     [f.Empty.CODE_NOT_EMPTY])
        # etc.

    def test_fail_non_collection(self):
        """
        The incoming value does not have a length.
        """
        # The Filter can't determine the length of this object, so it
        # assumes that it is not empty.
        self.assertFilterErrors(object(), [f.Empty.CODE_NOT_EMPTY])

    def test_zero_is_not_empty(self):
        """
        PHP developers take note!
        """
        self.assertFilterErrors(0, [f.Empty.CODE_NOT_EMPTY])

    def test_false_is_not_empty(self):
        """
        The boolean value ``False`` is NOT considered empty because it
        represents SOME kind of value.
        """
        self.assertFilterErrors(False, [f.Empty.CODE_NOT_EMPTY])


class MaxLengthTestCase(BaseFilterTestCase):
    filter_type = f.MaxLength

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use ``Required | MaxLength`` if you want to reject null values.
        """
        self.assertFilterPasses(
            self._filter(None, max_length=0),
        )

    def test_pass_short(self):
        """
        The incoming value is shorter than the max length.
        """
        self.assertFilterPasses(
            self._filter('Hello', max_length=6),
        )

    def test_pass_max_length(self):
        """
        The incoming value has the max allowed length.
        """
        self.assertFilterPasses(
            self._filter('World', max_length=5),
        )

    def test_fail_long(self):
        """
        The incoming value is longer than the max length.
        """
        self.assertFilterErrors(
            self._filter('Goodbye', max_length=5),
            [f.MaxLength.CODE_TOO_LONG],
        )

    def test_multi_byte_characters(self):
        """
        Multibyte characters are treated differently depending on
        whether you pass in a unicode or a byte string.
        """
        # "Hello world" in Chinese:
        decoded_value   = '\u4f60\u597d\u4e16\u754c'
        encoded_value   = decoded_value.encode('utf-8')


        # The string version of the string contains 4 code points.
        self.assertFilterPasses(
            self._filter(decoded_value, max_length=4),
        )

        # The bytes version of the string contains 12 bytes.
        self.assertFilterErrors(
            self._filter(encoded_value, max_length=4),
            [f.MaxLength.CODE_TOO_LONG],
        )

    def test_pass_short_collection(self):
        """
        The incoming value is a collection with length less than or
        equal to the max length.
        """
        self.assertFilterPasses(
            self._filter(['foo', 'bar', 'baz', 'luhrmann'], max_length=4),
        )

        self.assertFilterPasses(
            self._filter({'foo': 'bar', 'baz': 'luhrmann'}, max_length=3),
        )

        self.assertFilterPasses(
            self._filter(Lengthy(4), max_length=4),
        )

        # etc.

    def test_fail_long_collection(self):
        """
        The incoming value is a collection with length greater than the
        max length.
        """
        self.assertFilterErrors(
            self._filter(['foo', 'bar', 'baz', 'luhrmann'], max_length=3),
            [f.MaxLength.CODE_TOO_LONG],
        )

        self.assertFilterErrors(
            self._filter({'foo': 'bar', 'baz': 'luhrmann'}, max_length=1),
            [f.MaxLength.CODE_TOO_LONG],
        )

        self.assertFilterErrors(
            self._filter(Lengthy(4), max_length=3),
            [f.MaxLength.CODE_TOO_LONG],
        )

        # etc.


class MinLengthTestCase(BaseFilterTestCase):
    filter_type = f.MinLength

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use `Required | MinLength` if you want to reject null values.
        """
        self.assertFilterPasses(
            self._filter(None, min_length=5),
        )

    def test_pass_long(self):
        """
        The incoming value has length greater than the minimum value.
        """
        self.assertFilterPasses(
            self._filter('Hello', min_length=2),
        )

    def test_pass_min_length(self):
        """
        The incoming value has length equal to the minimum value.
        """
        self.assertFilterPasses(
            self._filter('World', min_length=5),
        )

    def test_fail_short(self):
        """
        The incoming value has length less than the minimum value.
        """
        self.assertFilterErrors(
            self._filter('Goodbye', min_length=10),
            [f.MinLength.CODE_TOO_SHORT],
        )

    def test_multi_byte_characters(self):
        """
        Multibyte characters are treated differently depending on
        whether you pass in a unicode or a byte string.
        """
        # "Hello world" in Chinese:
        decoded_value   = '\u4f60\u597d\u4e16\u754c'
        encoded_value   = decoded_value.encode('utf-8')

        # The string version of the string contains 4 code points.
        self.assertFilterErrors(
            self._filter(decoded_value, min_length=12),
            [f.MinLength.CODE_TOO_SHORT],
        )

        # The bytes version of the string contains 12 bytes.
        self.assertFilterPasses(
            self._filter(encoded_value, min_length=12),
        )

    def test_pass_long_collection(self):
        """
        The incoming value is a collection with length greater than or
        equal to the minimum value.
        """
        self.assertFilterPasses(
            self._filter(['foo', 'bar', 'baz', 'luhrmann'], min_length=3),
        )

        self.assertFilterPasses(
            self._filter({'foo': 'bar', 'baz': 'luhrmann'}, min_length=1),
        )

        self.assertFilterPasses(
            self._filter(Lengthy(6), min_length=5),
        )

        # etc.

    def test_fail_short_collection(self):
        """
        The incoming value is a collection with length less than the
        minimum value.
        """
        self.assertFilterErrors(
            self._filter(['foo', 'bar', 'baz', 'luhrmann'], min_length=5),
            [f.MinLength.CODE_TOO_SHORT],
        )

        self.assertFilterErrors(
            self._filter({'foo': 'bar', 'baz': 'luhrmann'}, min_length=3),
            [f.MinLength.CODE_TOO_SHORT],
        )

        self.assertFilterErrors(
            self._filter(Lengthy(6), min_length=7),
            [f.MinLength.CODE_TOO_SHORT],
        )

        # etc.


class NoOpTestCase(BaseFilterTestCase):
    filter_type = f.NoOp

    def test_pass_any_value(self):
        """
        You can pass any value you want to a NoOp, and it will pass.
        """
        self.assertFilterPasses('supercalafragalisticexpialadoshus')


class NotEmptyTestCase(BaseFilterTestCase):
    filter_type = f.NotEmpty

    def test_pass_none(self):
        """
        By default, :py:class:`f.NotEmpty` will treat ``None`` as
        valid, just like every other filter.

        Unlike every other filter, however, the strategy for rejecting
        null values is a wee bit different, as we'll see in the next
        test.
        """
        self.assertFilterPasses(None)

    def test_fail_none(self):
        """
        You can configure the filter to reject null values.
        """
        self.assertFilterErrors(
            self._filter(None, allow_none=False),
            [f.NotEmpty.CODE_EMPTY],
        )

    def test_pass_non_empty_string(self):
        """
        The incoming value is a non-empty string.
        """
        self.assertFilterPasses('Hello, world!')

    def test_pass_non_empty_collection(self):
        """
        The incoming value is a collection with length > 0.
        """
        # The values in the collection may be empty, but the collection
        # itself is not.
        self.assertFilterPasses(['', '', ''])
        self.assertFilterPasses({'': ''})
        self.assertFilterPasses(Lengthy(1))
        # etc.

    def test_pass_non_collection(self):
        """
        The incoming value does not have a length.
        """
        self.assertFilterPasses(object())

    def test_fail_empty_string(self):
        """
        The incoming value is an empty string.
        """
        self.assertFilterErrors('', [f.NotEmpty.CODE_EMPTY])

    def test_fail_empty_collection(self):
        """
        The incoming value is a collection with length < 1.
        """
        self.assertFilterErrors([],         [f.NotEmpty.CODE_EMPTY])
        self.assertFilterErrors({},         [f.NotEmpty.CODE_EMPTY])
        self.assertFilterErrors(Lengthy(0), [f.NotEmpty.CODE_EMPTY])
        # etc.

    def test_zero_is_not_empty(self):
        """
        PHP developers take note!
        """
        self.assertFilterPasses(0)

    def test_false_is_not_empty(self):
        """
        The boolean value ``False`` is NOT considered empty because it
        represents SOME kind of value.
        """
        self.assertFilterPasses(False)


class OptionalTestCase(BaseFilterTestCase):
    filter_type = f.Optional

    def test_pass_none(self):
        """
        It'd be pretty silly to name a Filter "Optional" if it rejects
        ``None``, wouldn't it?
        """
        self.assertFilterPasses(None)

    def test_replace_none(self):
        """
        The default replacement value is ``None``, but you can change
        it to something else.
        """
        self.assertFilterPasses(
            self._filter(None, default='Hello, world!'),
            'Hello, world!',
        )

    def test_replace_empty_string(self):
        """
        The incoming value is an empty string.
        """
        self.assertFilterPasses(
            self._filter('', default='42'),
            '42',
        )

    def test_replace_empty_collection(self):
        """
        The incoming value is a collection with length < 1.
        """
        # By default, the Filter will replace empty values with `None`.
        self.assertFilterPasses([],         None)
        self.assertFilterPasses({},         None)
        self.assertFilterPasses(Lengthy(0), None)
        # etc.

    def test_pass_non_empty_string(self):
        """
        The incoming value is a non-empty string.
        """
        self.assertFilterPasses(
            self._filter('Goodbye, world!', default='fail')
        )

    def test_pass_non_empty_collection(self):
        """
        The incoming value is a collection with length > 0.
        """
        # The values inside the collection may be empty, but the
        # collection itself is not.
        self.assertFilterPasses(['', '', ''])
        self.assertFilterPasses({'': ''})
        self.assertFilterPasses(Lengthy(12))
        # etc.

    def test_pass_non_collection(self):
        """
        Any value that doesn't have a length is left alone.
        """
        self.assertFilterPasses(
            self._filter(object(), default='fail'),
        )

    def test_zero_is_not_empty(self):
        """
        PHP developers take note!
        """
        self.assertFilterPasses(
            self._filter(0, default='fail'),
        )

    def test_false_is_not_empty(self):
        """
        The boolean value ``False`` is NOT considered empty because it
        represents SOME kind of value.
        """
        self.assertFilterPasses(
            self._filter(False, default='fail'),
        )


class RequiredTestCase(BaseFilterTestCase):
    filter_type = f.Required

    def test_fail_none(self):
        """
        :py:class:`f.Required` is the only filter that does not allow
        null values.
        """
        self.assertFilterErrors(None, [f.Required.CODE_EMPTY])

    def test_pass_non_empty_string(self):
        """
        The incoming value is a non-empty string.
        """
        self.assertFilterPasses('Hello, world!')

    def test_pass_non_empty_collection(self):
        """
        The incoming value is a collection with length > 0.
        """
        # The values in the collection may be empty, but the collection
        # itself is not.
        self.assertFilterPasses(['', '', ''])
        self.assertFilterPasses({'': ''})
        self.assertFilterPasses(Lengthy(1))
        # etc.

    def test_pass_non_collection(self):
        """
        Any value that does not have a length is assumed to be not
        empty.
        """
        self.assertFilterPasses(object())

    def test_fail_empty_string(self):
        """
        The incoming value is an empty string.
        """
        self.assertFilterErrors('', [f.Required.CODE_EMPTY])

    def test_fail_empty_collection(self):
        """
        The incoming value is a collection with length < 1.
        """
        self.assertFilterErrors([],         [f.Required.CODE_EMPTY])
        self.assertFilterErrors({},         [f.Required.CODE_EMPTY])
        self.assertFilterErrors(Lengthy(0), [f.Required.CODE_EMPTY])
        # etc.

    def test_zero_is_not_empty(self):
        """
        PHP developers take note!
        """
        self.assertFilterPasses(0)

    def test_false_is_not_empty(self):
        """
        The boolean value ``False`` is NOT considered empty because it
        represents SOME kind of value.
        """
        self.assertFilterPasses(False)


class TypeTestCase(BaseFilterTestCase):
    filter_type = f.Type

    def test_pass_none(self):
        """
        ``None`` always passes this Filter.

        Use ``Required | Type`` if you want to reject null values.
        """
        self.assertFilterPasses(
            self._filter(None, allowed_types=text_type),
        )

    def test_pass_matching_type(self):
        """
        The incoming value has the expected type.
        """
        self.assertFilterPasses(
            self._filter('Hello, world!', allowed_types=text_type),
        )

    def test_fail_non_matching_type(self):
        """
        The incoming value does not have the expected type.
        """
        self.assertFilterErrors(
            self._filter(b'Not a string, sorry.', allowed_types=text_type),
            [f.Type.CODE_WRONG_TYPE],
        )

    def test_multiple_allowed_types(self):
        """
        You can configure the Filter to allow multiple types.
        """
        self.assertFilterPasses(
            self._filter('Hello, world!', allowed_types=(text_type, int)),
        )

        self.assertFilterPasses(
            self._filter(42, allowed_types=(text_type, int)),
        )

        self.assertFilterErrors(
            self._filter(b'Not a unicode.', allowed_types=(text_type, int)),
            [f.Type.CODE_WRONG_TYPE],
        )

    def test_pass_subclass_allowed(self):
        """
        The incoming value's type is a subclass of an allowed type.
        """
        self.assertFilterPasses(
            # bool is a subclass of int.
            self._filter(True, allowed_types=int),
        )

    def test_fail_subclass_not_allowed(self):
        """
        You can configure the Filter to require exact type matches.
        """
        self.assertFilterErrors(
            self._filter(True, allowed_types=int, allow_subclass=False),
            [f.Type.CODE_WRONG_TYPE],
        )

    def test_fail_types_are_not_instances(self):
        """
        The Filter checks that the value is an INSTANCE of its allowed
        type(s).  It will reject the type(s) themselves.
        """
        self.assertFilterErrors(
            self._filter(text_type, allowed_types=text_type),
            [f.Type.CODE_WRONG_TYPE],
        )



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from abc import ABCMeta, abstractmethod as abstract_method
from copy import copy
from typing import Any, Dict, Iterable, List, Mapping, Optional, Sequence, \
    Text, Tuple, Union, Type, Callable
from weakref import ProxyTypes, proxy

from six import binary_type, python_2_unicode_compatible, text_type, \
    with_metaclass

__all__ = [
    'BaseFilter',
    'FilterChain',
    'FilterCompatible',
    'FilterError',
    'Type',
]


FilterCompatible =\
    Optional[Union['BaseFilter', 'FilterMeta', Callable[[], 'BaseFilter']]]
"""
Used in PEP-484 type hints to indicate a value that can be normalized
into an instance of a :py:class:`filters.base.BaseFilter` subclass.
"""


class FilterMeta(ABCMeta):
    """
    Metaclass for filters.
    """
    # noinspection PyShadowingBuiltins
    def __init__(cls, what, bases=None, dict=None):
        super(FilterMeta, cls).__init__(what, bases, dict)

        if not hasattr(cls, 'templates'):
            cls.templates = {}

        # Copy error templates from base class to derived class, but
        # in the event of a conflict, preserve the derived class'
        # template.
        templates = {}
        for base in bases:
            if isinstance(base, FilterMeta):
                templates.update(base.templates)

        if templates:
            templates.update(cls.templates)
            cls.templates = templates

    def __or__(self, next_filter):
        # type: (FilterCompatible) -> FilterChain
        """
        Convenience alias for adding a Filter with default
        configuration to a FilterChain.

        E.g., the following statements do the same thing::

            Int | Max(32)   # FilterMeta.__or__
            Int() | Max(32) # Filter.__or__

        References:
          - http://stackoverflow.com/a/10773232
        """
        return FilterChain(self) | next_filter


@python_2_unicode_compatible
class BaseFilter(with_metaclass(FilterMeta)):
    """
    Base functionality for all Filters, macros, etc.
    """
    CODE_EXCEPTION = 'exception'

    templates = {
        CODE_EXCEPTION: 'An error occurred while processing this value.',
    }

    def __init__(self):
        super(BaseFilter, self).__init__()

        self._parent = None # type: BaseFilter
        self._handler = None # type: BaseInvalidValueHandler
        self._key = None # type: Text

        #
        # Indicates whether the Filter detected any invalid values.
        # It gets reset every time `apply` gets called.
        #
        # Note that this attribute is intended to be used internally;
        # external code should instead interact with invalid value
        # handlers such as LogHandler and MemoryHandler.
        #
        # References:
        #   - :py:mod:`importer.core.filters.handlers`
        #
        self._has_errors = False

    # noinspection PyProtectedMember
    @classmethod
    def __copy__(cls, the_filter):
        # type: (BaseFilter) -> BaseFilter
        """
        Creates a shallow copy of the object.
        """
        new_filter = type(the_filter)() # type: BaseFilter

        new_filter._parent  = the_filter._parent
        new_filter._key     = the_filter._key
        new_filter._handler = the_filter._handler

        return new_filter

    def __or__(self, next_filter):
        # type: (FilterCompatible) -> FilterChain
        """
        Chains another filter with this one.
        """
        normalized = self.resolve_filter(next_filter)

        if normalized:
            #
            # Officially, we should do this:
            # return ``FilterChain(self) | next_filter``
            #
            # But that wastes some CPU cycles by creating an extra
            # FilterChain instance that gets thrown away almost
            # immediately. It's a bit faster just to create a single
            # FilterChain instance and modify it in-place.
            #
            # noinspection PyProtectedMember
            return FilterChain(self)._add(next_filter)
        else:
            return self if isinstance(self, FilterChain) else FilterChain(self)

    def __str__(self):
        """
        Returns a string representation of the Filter.

        Note that the output of this method does not necessarily match
        the signature of the Filter's ``__init__`` method; rather,
        its purpose is to provide a snapshot of critical parts of
        the Filter's configuration for e.g., troubleshooting
        purposes.
        """
        return '{type}()'.format(
            type = type(self).__name__,
        )

    @property
    def parent(self):
        # type: () -> Optional[BaseFilter]
        """
        Returns the parent Filter.
        """
        # Make sure `self._parent` hasn't gone away.
        try:
            self._parent.__class__
        except ReferenceError:
            return None

        return self._parent

    @parent.setter
    def parent(self, parent):
        # type: (BaseFilter) -> None
        """
        Sets the parent Filter.
        """
        # Create a weakref to the parent Filter to prevent annoying the
        # garbage collector.
        self._parent = (
            (parent if isinstance(parent, ProxyTypes) else proxy(parent))
                if parent
                else None
        )

    @property
    def key(self):
        # type: () -> Text
        """
        Returns the key associated with this filter.
        """
        return self._make_key(self._key_parts)

    @key.setter
    def key(self, key):
        # type: (Text) -> None
        """
        Sets the key associated with this filter.
        """
        self._key = key

    def sub_key(self, sub_key):
        # type: (Text) -> Text
        """
        Returns a copy of this filter's key with an additional sub-key
        appended.
        """
        return self._make_key(self._key_parts + [sub_key])

    @property
    def _key_parts(self):
        # type: () -> List[Text]
        """
        Assembles each key part in the filter hierarchy.
        """
        key_parts = []

        # Iterate up the parent chain and collect key parts.
        # Alternatively, we could just get ``self.parent._key_parts``,
        # but that is way less efficient.
        parent = self
        while parent:
            # As we move up the chain, push key parts onto the front of
            # the path (otherwise the key parts would be in reverse
            # order).
            key_parts.insert(0, parent._key)
            parent = parent.parent

        return key_parts

    @property
    def handler(self):
        # type: () -> BaseInvalidValueHandler
        """
        Returns the invalid value handler for the filter.
        """
        if self._handler is None:
            # Attempt to return the parent filter's handler...
            try:
                return self.parent.handler
            except AttributeError:
                #
                # ... unless this filter has no parent, in which case
                # it should use the default.
                #
                # Note that we do not set ``self._handler``, in case
                # the filter later gets added to e.g., a FilterChain
                # that has a different invalid value handler set.
                #
                return ExceptionHandler()

        return self._handler

    @handler.setter
    def handler(self, handler):
        # type: (BaseInvalidValueHandler) -> None
        """
        Sets the invalid value handler for the filter.
        """
        self._handler = handler

    def set_handler(self, handler):
        # type: (BaseInvalidValueHandler) -> BaseFilter
        """
        Cascading method for setting the filter's invalid value
        handler.
        """
        self.handler = handler
        return self

    def apply(self, value):
        """
        Applies the filter to a value.
        """
        self._has_errors = False

        try:
            return self._apply_none() if value is None else self._apply(value)
        except Exception as e:
            return self._invalid_value(value, e, exc_info=True)

    @abstract_method
    def _apply(self, value):
        """
        Applies filter-specific logic to a value.

        Note:  It is safe to assume that ``value`` is not ``None`` when
        this method is invoked.
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )

    def _apply_none(self):
        """
        Applies filter-specific logic when the value is ``None``.
        """
        return None

    def _filter(self, value, filter_chain, sub_key=None):
        # type: (Any, FilterCompatible) -> Any
        """
        Applies another filter to a value in the same context as the
        current filter.

        :param sub_key:
            Appended to the ``key`` value in the error message context
            (used by complex filters).
        """
        filter_chain = self.resolve_filter(filter_chain, parent=self, key=sub_key)

        # In rare cases, ``filter_chain`` may be ``None``.
        # :py:meth:`filters.complex.FilterMapper.__init__`
        if filter_chain:
            try:
                filtered = filter_chain.apply(value)
            except Exception as e:
                return self._invalid_value(value, e, exc_info=True)
            else:
                self._has_errors = self._has_errors or filter_chain._has_errors
                return filtered
        else:
            return value

    def _invalid_value(
            self,
            value,
            reason,
            replacement     = None,
            exc_info        = False,
            context         = None,
            sub_key         = None,
            template_vars   = None,
    ):
        # type: (Any, Union[Text, Exception], Any, bool, Optional[dict], Optional[Text], Optional[dict]) -> Any
        """
        Handles an invalid value.

        This method works as both a logging method and an exception
        handler.

        :param replacement:
            The replacement value to use instead.

        :param sub_key:
            Appended to the ``key`` value in the error message context
            (used by complex filters).

        :return:
            Replacement value to use instead of the invalid value
            (usually ``None``).
        """
        handler = self.handler

        if isinstance(reason, FilterError):
            # FilterErrors should be sent directly to the handler.
            # This allows complex Filters to properly catch and handle
            # FilterErrors raised by the Filters they control.
            return handler.handle_invalid_value(
                message     = text_type(reason),
                exc_info    = True,
                context     = getattr(reason, 'context', {}),
            )

        self._has_errors = True

        if not context:
            context = {}

        context['value']        = value
        context['filter']       = text_type(self)
        context['key']          = self.sub_key(sub_key)
        context['replacement']  = replacement

        if not template_vars:
            template_vars = {}

        template_vars.update(context)

        if isinstance(reason, Exception):
            # Store the error code in the context so that the caller
            # can identify the error type without having to parse the
            # rendered error message template.
            context['code'] = self.CODE_EXCEPTION

            # Store exception details in the context so that they are
            # accessible to devs but hidden from end users.
            # Note that the traceback gets processed separately,
            context['exc'] = '[{mod}.{cls}] {msg}'.format(
                mod = type(reason).__module__,
                cls = type(reason).__name__,
                msg = text_type(reason),
            )

            # Add the context to the exception object so that loggers
            # can use it.
            if not hasattr(reason, 'context'):
                reason.context = {}
            reason.context.update(context)

            handler.handle_exception(
                message = self._format_message(context['code'], template_vars),
                exc     = reason,
            )
        else:
            # Store the error code in the context so that the caller
            # can identify the error type without having to parse the
            # rendered error message template.
            context['code'] = reason

            handler.handle_invalid_value(
                message     = self._format_message(reason, template_vars),
                exc_info    = exc_info,
                context     = context,
            )

        return replacement

    def _format_message(self, key, template_vars):
        # type: (Text, Dict[Text, Text]) -> Text
        """
        Formats a message for the invalid value handler.
        """
        return self.templates[key].format(**template_vars)

    @classmethod
    def resolve_filter(cls, the_filter, parent=None, key=None):
        # type: (FilterCompatible, Optional[BaseFilter], Optional[Text]) -> Optional[FilterChain]
        """
        Converts a filter-compatible value into a consistent type.
        """
        if the_filter is not None:
            if isinstance(the_filter, BaseFilter):
                resolved = the_filter

            elif callable(the_filter):
                resolved = cls.resolve_filter(the_filter())

            # Uhh... hm.
            else:
                raise TypeError(
                    '{type} {value!r} is not '
                    'compatible with {target}.'.format(
                        type    = type(the_filter).__name__,
                        value   = the_filter,
                        target  = cls.__name__,
                    ),
                )

            if parent:
                resolved.parent = parent

            if key:
                resolved.key = key

            return resolved

    @staticmethod
    def _make_key(key_parts):
        # type: (Iterable[Text]) -> Text
        """
        Assembles a dotted key value from its component parts.
        """
        return '.'.join(filter(None, key_parts))


@python_2_unicode_compatible
class FilterChain(BaseFilter):
    """
    Allows you to chain multiple filters together so that they are
    treated as a single filter.
    """
    def __init__(self, start_filter=None):
        # type: (FilterCompatible) -> None
        super(FilterChain, self).__init__()

        self._filters = [] # type: List[BaseFilter]

        self._add(start_filter)

    def __str__(self):
        return '{type}({filters})'.format(
            type    = type(self).__name__,
            filters = ' | '.join(map(text_type, self._filters)),
        )

    def __or__(self, next_filter):
        # type: (FilterCompatible) -> FilterChain
        """
        Chains a filter with this one.

        This method creates a new FilterChain object without modifying
        the current one.
        """
        resolved = self.resolve_filter(next_filter)

        if resolved:
            new_chain = copy(self) # type: FilterChain
            new_chain._add(next_filter)
            return new_chain
        else:
            return self

    # noinspection PyProtectedMember
    @classmethod
    def __copy__(cls, the_filter):
        # type: (FilterChain) -> FilterChain
        """
        Creates a shallow copy of the object.
        """
        new_filter = super(FilterChain, cls).__copy__(the_filter) # type: FilterChain
        new_filter._filters = the_filter._filters[:]
        return new_filter

    def _add(self, next_filter):
        # type: (FilterCompatible) -> FilterChain
        """
        Adds a Filter to the collection directly.
        """
        resolved = self.resolve_filter(next_filter, parent=self)
        if resolved:
            self._filters.append(resolved)

        return self

    def _apply(self, value):
        for f in self._filters:
            value = self._filter(value, f)

            # FilterChains stop at the first sign of trouble.
            # This is important because FilterChains have to behave
            # consistently regardless of whether the invalid value
            # handler raises an exception.
            if self._has_errors:
                break

        return value

    def _apply_none(self):
        return self._apply(None)


class BaseInvalidValueHandler(with_metaclass(ABCMeta)):
    """
    Base functionality for classes that handle invalid values.
    """
    @abstract_method
    def handle_invalid_value(self, message, exc_info, context):
        # type: (Text, bool, dict) -> Any
        """
        Handles an invalid value.

        :param message:
            Error message.

        :param exc_info:
            Whether to include output from :py:func:``sys.exc_info``.

        :param context:
            Additional context values for the error.
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )

    def handle_exception(self, message, exc):
        # type: (Text, Exception) -> Any
        """
        Handles an uncaught exception.
        """
        return self.handle_invalid_value(
            message     = message,
            exc_info    = True,
            context     = getattr(exc, 'context', {}),
        )


class FilterError(ValueError):
    """
    Indicates that a parsed value could not be filtered because the
    value was invalid.
    """
    def __init__(self, *args, **kwargs):
        """
        Provides a container to include additional variables and other
        information to help troubleshoot errors.
        """
        # Exception kwargs are deprecated in Python 3, but keeping them
        # around for compatibility with Python 2.
        # noinspection PyArgumentList
        super(FilterError, self).__init__(*args, **kwargs)
        self.context = {}


class ExceptionHandler(BaseInvalidValueHandler):
    """
    Invalid value handler that raises an exception.
    """
    def handle_invalid_value(self, message, exc_info, context):
        error = FilterError(message)
        error.context = context
        raise error


# Used by :py:meth:`Type.__init__` to inject JSON data types in place of
# Python type names in error messages.
JSON_ALIASES = {
    # Builtins
    bool:           'Boolean',
    dict:           'Object',
    float:          'Number',
    int:            'Number',
    list:           'Array',

    # Compat
    binary_type:    'String',
    text_type:      'String',

    # Typing
    Mapping:        'Array',
    Sequence:       'Array',
}

# This filter is used extensively by other filters.
# To avoid lots of needless "circular import" hacks, we'll put it in
# the base module.
@python_2_unicode_compatible
class Type(BaseFilter):
    """
    Checks the type of a value.
    """
    CODE_WRONG_TYPE = 'wrong_type'

    templates = {
        CODE_WRONG_TYPE:
            '{incoming} is not valid (allowed types: {allowed}).',
    }

    def __init__(self, allowed_types, allow_subclass=True, aliases=None):
        # type: (Union[type, Tuple[type]], bool, Optional[Mapping[type, Text]]) -> None
        """
        :param allowed_types:
            The type (or types) that incoming values are allowed to
            have.

        :param allow_subclass:
            Whether to allow subclasses when checking for type matches.

        :param aliases:
            Aliases to use for type names in error messages.

            This is useful for providing more context- appropriate
            names to end users and/or masking native Python type names.
        """
        super(Type, self).__init__()

        # A pinch of syntactic sugar.
        self.allowed_types  = (
            allowed_types
                if isinstance(allowed_types, tuple)
                else (allowed_types,)
        )
        self.allow_subclass = allow_subclass

        self.aliases = aliases or {}

    def __str__(self):
        return (
            '{type}({allowed_types}, '
            'allow_subclass={allow_subclass!r})'.format(
                type            = type(self).__name__,
                allowed_types   = self.get_allowed_type_names(aliased=False),
                allow_subclass  = self.allow_subclass,
            )
        )

    def _apply(self, value):
        valid = (
            isinstance(value, self.allowed_types)
                if self.allow_subclass
                else (type(value) in self.allowed_types)
        )

        if not valid:
            return self._invalid_value(
                value   = value,
                reason  = self.CODE_WRONG_TYPE,

                template_vars = {
                    'incoming': self.get_type_name(type(value)),
                    'allowed':  self.get_allowed_type_names(),
                },
            )

        return value

    def get_allowed_type_names(self, aliased=True):
        # type: (bool) -> Text
        """
        Returns a string with all the allowed types.
        """
        # Note that we cast as a set in the middle, to ferret out
        # duplicates.
        return ', '.join(sorted({
            self.get_type_name(t, aliased)
                for t in self.allowed_types
        }))

    def get_type_name(self, type_, aliased=True):
        # type: (type, bool) -> Text
        """
        Returns the name of the specified type.
        """
        return (
            (self.aliases.get(type_) or type_.__name__)
                if aliased
                else type_.__name__
        )
